# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Optional, List
import logging

from .formatters import TTYFormatter, DaemonFormatter, auto_formatter
from .pager import AutoPager, enable_auto_paging, PagerError

__all__ = [
    TTYFormatter,
    DaemonFormatter,
    auto_formatter,
    AutoPager,
    enable_auto_paging,
    PagerError,
]


def setup_logger(
    debug: bool = False,
    log_flags: Optional[List[str]] = None,
    debug_flags: Optional[List[str]] = None,
    formatter: logging.Formatter = auto_formatter(),
    component: Optional[str] = None,
) -> None:
    """Setup root logger debug level, debug flags and formatter."""
    if debug:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.INFO

    if log_flags is None:
        log_flags = []
    if debug_flags is None:
        debug_flags = []

    root_logger = logging.getLogger()
    root_logger.setLevel(logging_level)
    handler = logging.StreamHandler()
    handler.setLevel(logging_level)
    formatter = formatter(debug, component)
    handler.setFormatter(formatter)

    # filter out all libs logs not enabled in flags
    def custom_filter(record):
        component = record.name.split(".")[0]
        if record.levelno == logging.DEBUG:
            if "ALL" in debug_flags or component in debug_flags:
                return 1
            return 0
        if "ALL" in log_flags or component in log_flags:
            return 1
        return 0

    handler.addFilter(custom_filter)
    root_logger.addHandler(handler)


def enforce_debug(
    log_flags: Optional[List[str]] = None,
    debug_flags: Optional[List[str]] = None,
) -> None:
    """Enforce root logger debug level and debug flags."""
    root_logger = logging.getLogger()
    root_logger.setLevel(level=logging.DEBUG)

    if log_flags is None:
        log_flags = []
    if debug_flags is None:
        debug_flags = []

    # filter out all libs logs not enabled in flags
    def custom_filter(record):
        component = record.name.split(".")[0]
        if record.levelno == logging.DEBUG:
            if "ALL" in debug_flags or component in debug_flags:
                return 1
            return 0
        if "ALL" in log_flags or component in log_flags:
            return 1
        return 0

    # For all handlers, set log level, enable debug in formatter and replace filter.
    for handler in root_logger.handlers:
        handler.setLevel(logging.DEBUG)
        handler.formatter.debug = True
        for filter in handler.filters:
            handler.removeFilter(filter)
            handler.addFilter(custom_filter)



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Optional, List, Tuple
from pathlib import Path
import logging

try:
    import ldap
    import ldap.filter
except ImportError as err:
    raise ImportError("python-ldap is required for RFL LDAP Authentication") from err

from .user import AuthenticatedUser
from .errors import LDAPAuthenticationError


logger = logging.getLogger(__name__)


class LDAPAuthentifier:
    def __init__(
        self,
        uri,
        user_base: str,
        group_base: str,
        user_class: str = "posixAccount",
        user_name_attribute: str = "uid",
        user_fullname_attribute: str = "cn",
        user_primary_group_attribute: str = "gidNumber",
        group_name_attribute: str = "cn",
        group_object_classes: Optional[List[str]] = None,
        cacert: Optional[Path] = None,
        starttls: bool = False,
        bind_dn: Optional[str] = None,
        bind_password: Optional[str] = None,
        restricted_groups: Optional[List[str]] = None,
        lookup_user_dn: bool = False,
        lookup_as_user: Optional[bool] = True,
    ):
        self.uri = uri
        self.cacert = cacert
        self.user_base = user_base
        self.user_class = user_class
        self.group_base = group_base
        self.user_name_attribute = user_name_attribute
        self.user_fullname_attribute = user_fullname_attribute
        self.user_primary_group_attribute = user_primary_group_attribute
        self.group_name_attribute = group_name_attribute
        if group_object_classes is None:
            # Standard RFC 2307 (aka. NIS) schema has group entries with posixGroup
            # structural class. RFC 2307 bis schema has group entries with generally at
            # least groupOfNames class.
            group_object_classes = ["posixGroup", "groupOfNames"]
        self.group_object_classes = group_object_classes
        self.starttls = starttls
        self.bind_dn = bind_dn
        self.bind_password = bind_password
        self.restricted_groups = restricted_groups
        self.lookup_user_dn = lookup_user_dn
        # The lookup_as_user attribute is a boolean to control which bind dn is used to
        # retrieve user information and user groups after successful authentication.
        #
        # When True (default), LDAP connection is kept after authentication to use user
        # permissions.
        #
        # When False, LDAP connection is closed, another LDAP connection is opened. This
        # connection is binded with service bind dn and password when defined. When not
        # defined, user information and groups are retrieved anonymously.
        #
        # When the lookup_as_user argument is not defined, it is enabled when either
        # bind dn or password are not defined, and disabled otherwise. When argument
        # value is True or False, this value is directly used.
        if lookup_as_user is None:
            if self.bind_dn is None or self.bind_password is None:
                self.lookup_as_user = True
            else:
                self.lookup_as_user = False
        else:
            self.lookup_as_user = lookup_as_user

    def connection(self):
        connection = ldap.initialize(self.uri.geturl())
        # SSL/TLS setup
        if self.uri.geturl().startswith("ldaps") or self.starttls:
            connection.protocol_version = ldap.VERSION3
            # Force server certificate validation
            connection.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_DEMAND)
            # For LDAPS and STARTTLS, libldap require the path to CA certificates to be
            # defined to authentication server certificate. If the cacert option is
            # defined, use it else use default system CA certificates directory defined
            # in OpenSSL library.
            if self.cacert is None:
                import ssl

                logger.debug(
                    "Using default system OpenSSL CA certificate directory to "
                    "authenticate server"
                )
                connection.set_option(
                    ldap.OPT_X_TLS_CACERTDIR,
                    ssl.get_default_verify_paths().openssl_capath,
                )
            else:
                logger.debug(
                    "Using CA certification %s to authenticate server", self.cacert
                )
                connection.set_option(ldap.OPT_X_TLS_CACERTFILE, str(self.cacert))
            # Force libldap to create a new SSL context
            try:
                connection.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
            except ValueError as err:
                raise LDAPAuthenticationError(
                    "LDAP connection option value error"
                ) from err
        if self.starttls:
            try:
                logger.debug("Using STARTTLS to initialize TLS connection")
                connection.start_tls_s()
            except ldap.CONNECT_ERROR as err:
                raise LDAPAuthenticationError(
                    f"Unable to connect to LDAP server with STARTTLS: {str(err)}"
                ) from err
            except ldap.SERVER_DOWN as err:
                raise LDAPAuthenticationError(
                    f"LDAP server {self.uri.geturl()} is unreachable"
                ) from err
        return connection

    def _get_user_info(
        self, connection: ldap.ldapobject.LDAPObject, user_dn: str
    ) -> Tuple[str, int]:
        """Return fullname and primary group number from the provided user DN."""
        search_filter = f"(objectClass={self.user_class})"
        try:
            results = connection.search_s(
                user_dn,
                ldap.SCOPE_BASE,
                search_filter,
                [self.user_fullname_attribute, self.user_primary_group_attribute],
            )
        except ldap.NO_SUCH_OBJECT as err:
            raise LDAPAuthenticationError(f"Unable to find user DN {user_dn}") from err
        logger.debug(
            "LDAP search base: %s, scope: base filter: %s, results: %s",
            user_dn,
            search_filter,
            str(results),
        )
        if not len(results):
            raise LDAPAuthenticationError(
                f"User not found in LDAP with class {self.user_class}"
            )
        try:
            fullname = results[0][1][self.user_fullname_attribute][0].decode()
        except KeyError as err:
            raise LDAPAuthenticationError(
                f"Unable to extract user full name with {self.group_name_attribute} "
                "attribute from user entries"
            ) from err
        try:
            gid = int(results[0][1][self.user_primary_group_attribute][0])
        except KeyError:
            logger.warning(
                "Unable to extract user primary group with %s attribute from user "
                "entry",
                self.user_primary_group_attribute,
            )
            gid = None
        return fullname, gid

    def _get_groups(
        self,
        connection: ldap.ldapobject.LDAPObject,
        user_name: str,
        user_dn: str,
        gid: Optional[int],
    ) -> List[str]:
        """Return the list of groups whose provided user is member, including its
        primary group ID. This function supports both RFC 2307 (aka. NIS schema) and
        RFC 2307bis schema."""
        # In standard RFC 2307 (aka. NIS) schema, group members are declared with
        # memberUid attributes (with user cn as values).
        #
        # In RFC 2307 bis schema, group members are declared with member attributes
        # (with full user dn as values).
        #
        # In both cases, user primary group declared in user entry (gid argument) must
        # not be forgiven if defined.
        object_class_filter = "".join(
            [
                f"(objectClass={object_class})"
                for object_class in self.group_object_classes
            ]
        )
        gid_filter = f"(gidNumber={gid})" if gid is not None else ""
        search_filter = (
            "(&"
            f"(|{object_class_filter})"
            f"(|(memberUid={ldap.filter.escape_filter_chars(user_name)})"
            f"(member={ldap.filter.escape_filter_chars(user_dn)}){gid_filter}))"
        )
        try:
            results = connection.search_s(
                self.group_base,
                ldap.SCOPE_SUBTREE,
                search_filter,
                [self.group_name_attribute],
            )
        except ldap.NO_SUCH_OBJECT as err:
            raise LDAPAuthenticationError(
                f"Unable to find group base {self.group_base}"
            ) from err
        logger.debug(
            "LDAP search base: %s, scope: subtree, filter: %s, results: %s",
            self.group_base,
            search_filter,
            str(results),
        )
        if not len(results):
            logger.warning(
                "Unable to find groups in LDAP for user %s%s",
                user_name,
                f" or gidNumber {gid}" if gid is not None else "",
            )
        try:
            return [
                result[1][self.group_name_attribute][0].decode() for result in results
            ]
        except KeyError as err:
            raise LDAPAuthenticationError(
                f"Unable to extract group name with {self.group_name_attribute} "
                "attribute from group entries"
            ) from err

    def _in_restricted_groups(self, groups: List[str]):
        """Return False if restricted groups are set and none of the groups in argument
        matches the restricted groups. If either restricted groups are unset or any of
        the groups in argument is in restricted group, True is returned."""
        return not (
            self.restricted_groups is not None
            and len(self.restricted_groups)
            and not any([group in self.restricted_groups for group in groups])
        )

    def _bind(self, connection):
        """Bind connection with provided dn/password. Binding is skipped if bind_dn
        attribute is None. LDAPAuthentication is raised if either bind_password
        attribute is None (while bind_dn is defined) or LDAP server returns invalid
        credential error."""
        if self.bind_dn is None:
            return
        logger.debug("Using DN %s to bind to LDAP directory", self.bind_dn)
        try:
            assert self.bind_password is not None
        except AssertionError as err:
            raise LDAPAuthenticationError(
                f"Password to authenticate with bind DN {self.bind_dn} is required"
            ) from err
        try:
            connection.simple_bind_s(self.bind_dn, self.bind_password)
        except ldap.INVALID_CREDENTIALS as err:
            raise LDAPAuthenticationError("Invalid bind DN or password") from err
        except ldap.SERVER_DOWN as err:
            raise LDAPAuthenticationError(
                f"LDAP server {self.uri.geturl()} is unreachable"
            ) from err

    def _lookup_user_dn(self, user):
        """Return the DN of the given user in user base subtree. If lookup_user_dn flag
        is False, just concatenate of the user name attribute, the provided user name
        and the user base. Else, bind to LDAP directory with bind_{dn,password} if
        provided and search for the user DN in subtree scope. Check there is only one
        matching result and return the DN. Raise LDAPAuthenticationError when user base
        not found and when number of results is not one."""

        # If lookup_user_dn flag is disabled, just return the concatenation of the user
        # name attribute, the provided user name and the user base.
        if not self.lookup_user_dn:
            return f"{self.user_name_attribute}={user},{self.user_base}"

        connection = self.connection()

        # Bind with bind_{dn,password} if defined.
        self._bind(connection)

        search_filter = (
            f"(&(objectClass={self.user_class})"
            f"({self.user_name_attribute}={ldap.filter.escape_filter_chars(user)}))"
        )
        try:
            results = connection.search_s(
                self.user_base,
                ldap.SCOPE_SUBTREE,
                search_filter,
                [self.user_name_attribute],
            )
        except ldap.SERVER_DOWN as err:
            raise LDAPAuthenticationError(
                f"LDAP server {self.uri.geturl()} is unreachable"
            ) from err
        except ldap.NO_SUCH_OBJECT as err:
            raise LDAPAuthenticationError(
                f"Unable to find user base {self.user_base}"
            ) from err
        except ldap.OPERATIONS_ERROR as err:
            raise LDAPAuthenticationError(
                f"Operations error on user DN lookup: {err}"
            ) from err
        finally:
            connection.unbind_s()
        logger.debug(
            "LDAP search base: %s, scope: subtree, filter: %s, results: %s",
            self.group_base,
            search_filter,
            str(results),
        )
        # Check at least one result has been found or raise error
        if not len(results):
            logger.warning(
                "Unable to find user %s=%s in LDAP in base %s subtree",
                self.user_name_attribute,
                user,
                self.user_base,
            )
            raise LDAPAuthenticationError(
                f"Unable to find user {user} in base {self.user_base}"
            )
        # Check no more than one result has been found or raise error
        if len(results) > 1:
            logger.warning(
                "Too many users found (%d) with %s=%s in LDAP in base %s subtree",
                len(results),
                self.user_name_attribute,
                user,
                self.user_base,
            )
            raise LDAPAuthenticationError(
                f"Too many users found ({len(results)}) with username {user} in base "
                f"{self.user_base}"
            )
        # Return DN of this unique result
        return results[0][0]

    def login(self, user: str, password: str) -> AuthenticatedUser:
        """Verify provided user/password are valid and return the corresponding
        AuthenticatedUser. Raise LDAPAuthenticationError if restricted groups are set
        and the user in not member of any of these groups."""
        fullname = None
        groups = None
        if user is None or password is None:
            raise LDAPAuthenticationError("Invalid authentication request")

        # Lookup user DN in user base.
        user_dn = self._lookup_user_dn(user)
        logger.debug("DN user for authentication of user %s: %s", user, user_dn)

        connection = self.connection()
        try:
            # Bind with user password to test their credentials
            connection.simple_bind_s(user_dn, password)

            # Unless lookup as user is enabled, close current connection with
            # authenticated user credentials, open another connection and bind with
            # service dn (if defined).
            if not self.lookup_as_user:
                logger.debug("Re-initialize LDAP connection to avoid lookup as user")
                connection.unbind_s()
                connection = self.connection()
                self._bind(connection)

            fullname, gid = self._get_user_info(connection, user_dn)
            groups = self._get_groups(connection, user, user_dn, gid)
        except ldap.SERVER_DOWN as err:
            raise LDAPAuthenticationError(
                f"LDAP server {self.uri.geturl()} is unreachable"
            ) from err
        except ldap.INVALID_CREDENTIALS as err:
            raise LDAPAuthenticationError("Invalid user or password") from err
        except ldap.UNWILLING_TO_PERFORM as err:
            raise LDAPAuthenticationError(
                f"LDAP server is unwilling to perform: {str(err)}"
            ) from err
        finally:
            connection.unbind_s()
        if not self._in_restricted_groups(groups):
            raise LDAPAuthenticationError(
                f"User {user} is not member of restricted groups"
            )
        return AuthenticatedUser(login=user, fullname=fullname, groups=groups)

    def _list_user_dn(self, connection):
        """Return list of all users name/pairs pairs in LDAP directory."""
        search_filter = f"(objectClass={self.user_class})"
        try:
            results = connection.search_s(
                self.user_base,
                ldap.SCOPE_SUBTREE,
                search_filter,
                [self.user_name_attribute],
            )
        except ldap.NO_SUCH_OBJECT as err:
            raise LDAPAuthenticationError(
                f"Unable to find user base {self.user_base}"
            ) from err
        except ldap.OPERATIONS_ERROR as err:
            raise LDAPAuthenticationError(
                f"Operations error on users search: {err}"
            ) from err
        logger.debug(
            "LDAP search base: %s, scope: subtree, filter: %s, results: %s",
            self.group_base,
            search_filter,
            str(results),
        )
        if not len(results):
            logger.warning(
                "Unable to find users in LDAP in base %s subtree",
                self.user_base,
            )
            # Return an empty list to avoid further processing.
            return []

        # Pick results where user name attribute is found. If attribute not found in
        # any result, raise LDAPAuthenticationError.
        user_name_attribute_found = False
        picked = []
        for result in results:
            if self.user_name_attribute not in result[1]:
                logger.warning(
                    "Unable to find %s from user entry %s",
                    self.user_name_attribute,
                    result[0],
                )
                continue
            user_name_attribute_found = True
            picked.append((result[1][self.user_name_attribute][0].decode(), result[0]))
        if not user_name_attribute_found:
            raise LDAPAuthenticationError(
                f"Unable to extract user {self.user_name_attribute} from user entries"
            )
        return picked

    def users(self, with_groups: bool = False) -> List[AuthenticatedUser]:
        """Return list of AuthenticatedUser available in LDAP directory. If with_groups
        is True, the groups attribute of the AuthenticatedUsers is also initialized with
        the list of their groups. If with_groups is True and restricted groups are set,
        all users whose groups do not match any of the restricted groups are
        discarded."""
        result = []
        connection = self.connection()

        # Bind with bind_{dn,password} if defined.
        self._bind(connection)

        try:
            for user, user_dn in self._list_user_dn(connection):
                fullname, gid = self._get_user_info(connection, user_dn)
                groups = []
                if with_groups:
                    groups = self._get_groups(connection, user, user_dn, gid)
                    # Skip the user if not member of any of the restricted groups
                    if not self._in_restricted_groups(groups):
                        logger.debug(
                            "Discarding user %s not member of restricted groups", user
                        )
                        continue
                result.append(
                    AuthenticatedUser(login=user, fullname=fullname, groups=groups)
                )
        except ldap.SERVER_DOWN as err:
            raise LDAPAuthenticationError(
                f"LDAP server {self.uri.geturl()} is unreachable"
            ) from err
        except ldap.UNWILLING_TO_PERFORM as err:
            raise LDAPAuthenticationError(
                f"LDAP server is unwilling to perform: {str(err)}"
            ) from err
        finally:
            connection.unbind_s()
        return result



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from pathlib import Path
import urllib
import ipaddress

import warnings
import logging

from .definition import (
    SettingsDefinition,
    SettingsSectionDefinition,
    SettingsParameterDefinition,
    SettingsDefinitionLoaderYaml,
)
from .loaders import RuntimeSettingsSiteLoader, RuntimeSettingsSiteLoaderIni
from .errors import SettingsOverrideError

logger = logging.getLogger(__name__)


class RuntimeSettingsSection:
    def __init__(self, section: SettingsSectionDefinition, origin: str):
        self._name = section.name
        self._origin = {}
        for parameter in section.parameters:
            setattr(self, parameter.name, parameter.default)
            self._origin[parameter.name] = origin


class RuntimeSettings:
    def __init__(self, definition: SettingsDefinition):
        # set to hold all sections names
        self._sections = set()
        self._definition = definition
        for section in definition.sections:
            o_section = type(
                f"{self.__class__.__name__}{section.name.capitalize()}",
                (RuntimeSettingsSection,),
                dict(),
            )(section, definition.loader.name)
            setattr(self, o_section._name, o_section)
            self._sections.add(o_section._name)

    def override(self, loader: RuntimeSettingsSiteLoader):
        # load site ini file and check it matches definition
        for section_name in loader.content.sections():
            if not self._definition.has_section(section_name):
                raise SettingsOverrideError(
                    f"Section {section_name} loaded in settings overrides is not "
                    "defined in settings definition"
                )
            section = self._definition.section(section_name)
            for parameter_name in loader.content[section.name]:
                if not section.has_parameter(parameter_name):
                    raise SettingsOverrideError(
                        f"Parameter {parameter_name} loaded in settings overrides is "
                        f"not defined in section {section.name} of settings definition"
                    )
                parameter = section.parameter(parameter_name)
                self._override_parameter(loader, section, parameter)

        # check required parameters
        self._check_required()

    def _check_required(self):
        """Verify that all parameters declared as required in settings definition are
        properly defined with a real value or raise SettingsOverrideError."""
        for section in self._definition.sections:
            for parameter in section.parameters:
                if (
                    parameter.required
                    and getattr(getattr(self, parameter.section), parameter.name)
                    is None
                ):
                    raise SettingsOverrideError(
                        f"Parameter {str(parameter)} is missing but required in "
                        "settings overrides"
                    )

    def _override_parameter(
        self,
        loader: RuntimeSettingsSiteLoader,
        section: SettingsSectionDefinition,
        parameter: SettingsParameterDefinition,
    ) -> None:
        """Load value from INI file, convert it to expected type, verify it is a valid
        choice and set corresponding object attribute. Special logic is applied for list
        to iterate over the items."""
        raw = loader.content[section.name].get(parameter.name)
        # Check if parameter is deprecated.
        if parameter.deprecated:
            # Check replacement parameter is not defined as well. If yes, ignore
            # deprecated parameter and log warning message.
            if (
                parameter.deprecated["section"] in loader.content
                and parameter.deprecated["parameter"]
                in loader.content[parameter.deprecated["section"]]
            ):
                logger.warning(
                    "Ignoring deprecated parameter %s since replacement parameter "
                    "[%s]>%s is also defined",
                    parameter,
                    parameter.deprecated["section"],
                    parameter.deprecated["parameter"],
                )
                return
            # Deprecated parameter is actually loaded, emit FutureWarning to alert end
            # user about the deprecation notice.
            warnings.warn(
                f"Parameter {parameter} is deprecated in favor of "
                f"[{parameter.deprecated['section']}]>"
                f"{parameter.deprecated['parameter']}",
                FutureWarning,
            )
        if parameter._type == "list":
            value = [
                self._load_parameter_value(parameter, parameter.content, raw)
                for raw in raw.strip().split("\n")
            ]
            for item in value:
                self._validate_choice(parameter, item)
        else:
            value = self._load_parameter_value(parameter, parameter._type, raw)
            self._validate_choice(parameter, value)
        setattr(getattr(self, section.name), parameter.name, value)
        getattr(self, section.name)._origin[parameter.name] = loader.name

    def _load_parameter_value(self, parameter, _type: str, raw: str):
        """Try to convert raw value to expected type or raise SettingsOverrideError."""
        if _type == "path":
            # Converting a string to a pathlib.Path basically never fails.
            return Path(raw)
        elif _type == "uri":
            # Converting a string to a urllib.ParseResult basically never fails.
            return urllib.parse.urlparse(raw)
        elif _type == "ip":
            try:
                return ipaddress.ip_address(raw)
            except ValueError as err:
                raise SettingsOverrideError(
                    f"Invalid ip address value '{raw}' for {parameter} in site "
                    "overrides"
                ) from err
        elif _type == "network":
            try:
                return ipaddress.ip_network(raw)
            except ValueError as err:
                raise SettingsOverrideError(
                    f"Invalid ip network value '{raw}' for {parameter} in site "
                    "overrides"
                ) from err
        elif _type == "int":
            try:
                return int(raw)
            except ValueError as err:
                raise SettingsOverrideError(
                    f"Invalid integer value '{raw}' for {parameter} in site overrides"
                ) from err
        elif _type == "float":
            try:
                return float(raw)
            except ValueError as err:
                raise SettingsOverrideError(
                    f"Invalid float value '{raw}' for {parameter} in site overrides"
                ) from err
        elif _type == "bool":
            if raw.lower() in ["1", "yes", "true", "on"]:
                return True
            elif raw.lower() in ["0", "no", "false", "off"]:
                return False
            else:
                raise SettingsOverrideError(
                    f"Invalid boolean value '{raw}' for {parameter} in site overrides"
                )
        else:
            # Return unmodified value
            return raw

    def _validate_choice(self, parameter, value):
        """Validate value is in parameter choices or raise SettingsOverrideError."""
        if parameter.choices is not None and value not in parameter.choices:
            raise SettingsOverrideError(
                f"Value {value} for parameter {parameter} in site overrides is "
                f"not one of possible choices {parameter.choices}"
            )

    def _dump_typed_value(self, value, _type) -> str:
        """Dump a specific value according to its type."""
        # Hide password
        if _type == "password":
            value = "•" * len(value)
        elif _type == "uri":
            value = value.geturl()
        elif _type in ["ip", "network"]:
            return str(value)
        return value

    def dump(self):
        """Print configuration parameters grouped by section, with the origin of their
        value (definition or site) on standard output."""
        for section_name in sorted(self._sections):
            print(f"[{section_name}]")
            section = getattr(self, section_name)
            for parameter_name, origin in section._origin.items():
                value = getattr(section, parameter_name)
                # Special character for none value.
                if value is None:
                    value = "∅"
                else:
                    # If list, print items comma-separated
                    if (
                        self._definition.section(section_name)
                        .parameter(parameter_name)
                        ._type
                        == "list"
                    ):
                        value = ", ".join(
                            [
                                self._dump_typed_value(
                                    item,
                                    self._definition.section(section_name)
                                    .parameter(parameter_name)
                                    .content,
                                )
                                for item in value
                            ]
                        )
                    else:
                        value = self._dump_typed_value(
                            value,
                            self._definition.section(section_name)
                            .parameter(parameter_name)
                            ._type,
                        )
                # Add tag for deprecated parameters
                tag = ""
                if (
                    self._definition.section(section_name)
                    .parameter(parameter_name)
                    .deprecated
                ):
                    tag = "[🪦]"
                print(
                    f"  {parameter_name}{tag}: {value} ({origin})",
                )

    def override_ini(self, path: Path) -> None:
        self.override(RuntimeSettingsSiteLoaderIni(path=path))

    @classmethod
    def yaml_definition(cls, path: Path):
        return cls(SettingsDefinition(SettingsDefinitionLoaderYaml(path=path)))



# Copyright (c) 2026 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
from unittest.mock import MagicMock, patch

import flask

try:
    from authlib.integrations.base_client import OAuthError

    from rfl.authentication.errors import OIDCAuthenticationError
    from rfl.authentication.oidc import OIDCClient
    from rfl.authentication.user import AuthenticatedUser
except ImportError:
    AUTHLIB_AVAILABLE = False
else:
    AUTHLIB_AVAILABLE = True


@unittest.skipUnless(AUTHLIB_AVAILABLE, "authlib is not installed")
class TestOIDCClient(unittest.TestCase):
    def setUp(self):
        self.app = flask.Flask(__name__)
        self.app.secret_key = "test-secret"

    def _make_client(self, **kwargs):
        defaults = {
            "issuer": "https://idp.example.com",
            "client_id": "client-id",
            "client_secret": "client-secret",
            "redirect_uri": "https://app.example.com/callback",
        }
        defaults.update(kwargs)
        with self.app.app_context():
            return OIDCClient(self.app, **defaults)

    @patch("rfl.authentication.oidc.OAuth")
    def test_init_registers_client(self, mock_oauth_cls):
        mock_oauth = mock_oauth_cls.return_value
        mock_client = MagicMock()
        mock_oauth.register.return_value = mock_client

        client = self._make_client(scope="openid email", cacert="/etc/ssl/ca.pem")

        mock_oauth_cls.assert_called_once_with(
            self.app,
            cache=None,
            fetch_token=None,
            update_token=None,
        )
        mock_oauth.register.assert_called_once_with(
            "rfl_oidc",
            client_id="client-id",
            client_secret="client-secret",
            server_metadata_url="https://idp.example.com/.well-known/openid-configuration",
            client_kwargs={
                "scope": "openid email",
                "verify": "/etc/ssl/ca.pem",
            },
        )
        self.assertIs(client._client, mock_client)

    @patch("rfl.authentication.oidc.OAuth")
    def test_init_with_pkce(self, mock_oauth_cls):
        mock_oauth_cls.return_value.register.return_value = MagicMock()

        self._make_client(pkce="S256")

        register_kwargs = mock_oauth_cls.return_value.register.call_args.kwargs
        self.assertEqual(
            register_kwargs["client_kwargs"]["code_challenge_method"],
            "S256",
        )

    @patch("rfl.authentication.oidc.OAuth")
    def test_init_without_pkce(self, mock_oauth_cls):
        mock_oauth_cls.return_value.register.return_value = MagicMock()

        self._make_client()

        register_kwargs = mock_oauth_cls.return_value.register.call_args.kwargs
        self.assertNotIn("code_challenge_method", register_kwargs["client_kwargs"])

    @patch("rfl.authentication.oidc.OAuth")
    def test_init_public_client(self, mock_oauth_cls):
        mock_oauth_cls.return_value.register.return_value = MagicMock()

        client = self._make_client(client_secret=None, pkce="S256")

        mock_oauth_cls.return_value.register.assert_called_once_with(
            "rfl_oidc",
            client_id="client-id",
            client_secret=None,
            server_metadata_url="https://idp.example.com/.well-known/openid-configuration",
            client_kwargs={
                "scope": "openid profile email",
                "verify": True,
                "code_challenge_method": "S256",
            },
        )
        self.assertIsNone(client.client_secret)

    @patch("rfl.authentication.oidc.OAuth")
    def test_init_public_client_empty_secret(self, mock_oauth_cls):
        mock_oauth_cls.return_value.register.return_value = MagicMock()

        client = self._make_client(client_secret="", pkce="S256")

        register_kwargs = mock_oauth_cls.return_value.register.call_args.kwargs
        self.assertIsNone(register_kwargs["client_secret"])
        self.assertIsNone(client.client_secret)

    @patch("rfl.authentication.oidc.OAuth")
    def test_init_public_client_requires_pkce(self, mock_oauth_cls):
        with self.assertRaisesRegex(
            OIDCAuthenticationError,
            "PKCE is required for public OIDC clients",
        ):
            self._make_client(client_secret=None, pkce=None)

        mock_oauth_cls.return_value.register.assert_not_called()

    @patch("rfl.authentication.oidc.OAuth")
    def test_redirect_uses_default_redirect_uri(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_response = MagicMock()
        mock_client.authorize_redirect.return_value = mock_response

        client = self._make_client()
        with self.app.test_request_context():
            response = client.redirect()

        mock_client.authorize_redirect.assert_called_once_with(
            redirect_uri="https://app.example.com/callback",
        )
        self.assertIs(response, mock_response)

    @patch("rfl.authentication.oidc.OAuth")
    def test_redirect_custom_redirect_uri(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client

        client = self._make_client()
        with self.app.test_request_context():
            client.redirect(redirect_uri="https://other.example/cb")

        mock_client.authorize_redirect.assert_called_once_with(
            redirect_uri="https://other.example/cb",
        )

    @patch("rfl.authentication.oidc.OAuth")
    def test_authenticate_returns_authenticated_user(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_access_token.return_value = {
            "userinfo": {
                "sub": "alice",
                "name": "Alice",
                "groups": ["users", "admins"],
            },
        }

        client = self._make_client()
        with self.app.test_request_context():
            user = client.authenticate()

        self.assertIsInstance(user, AuthenticatedUser)
        self.assertEqual(user.login, "alice")
        self.assertEqual(user.fullname, "Alice")
        self.assertEqual(user.groups, ["users", "admins"])

    @patch("rfl.authentication.oidc.OAuth")
    def test_authenticate_custom_claims(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_access_token.return_value = {
            "userinfo": {
                "uid": "bob",
                "displayName": "Bob",
                "memberOf": "operators",
            },
        }

        client = self._make_client(
            subject_claim="uid",
            fullname_claim="displayName",
            groups_claim="memberOf",
        )
        with self.app.test_request_context():
            user = client.authenticate()

        self.assertEqual(user.login, "bob")
        self.assertEqual(user.fullname, "Bob")
        self.assertEqual(user.groups, ["operators"])

    @patch("rfl.authentication.oidc.OAuth")
    def test_authenticate_missing_userinfo(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_access_token.return_value = {"access_token": "tok"}

        client = self._make_client()
        with self.app.test_request_context():
            with self.assertRaisesRegex(
                OIDCAuthenticationError,
                "validated userinfo",
            ):
                client.authenticate()

    @patch("rfl.authentication.oidc.OAuth")
    def test_authenticate_missing_subject(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_access_token.return_value = {
            "userinfo": {"name": "Alice"},
        }

        client = self._make_client()
        with self.app.test_request_context():
            with self.assertRaisesRegex(
                OIDCAuthenticationError,
                r"subject claim sub",
            ):
                client.authenticate()

    @patch("rfl.authentication.oidc.OAuth")
    def test_authenticate_groups_claim_disabled(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_access_token.return_value = {
            "userinfo": {"sub": "alice", "groups": ["admins"]},
        }

        client = self._make_client(groups_claim=None)
        with self.app.test_request_context():
            user = client.authenticate()

        self.assertEqual(user.groups, [])

    @patch("rfl.authentication.oidc.OAuth")
    def test_authenticate_restricted_groups_allowed(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_access_token.return_value = {
            "userinfo": {"sub": "alice", "groups": ["admins", "users"]},
        }

        client = self._make_client(restricted_groups=["admins"])
        with self.app.test_request_context():
            user = client.authenticate()

        self.assertEqual(user.login, "alice")

    @patch("rfl.authentication.oidc.OAuth")
    def test_authenticate_restricted_groups_rejected(self, mock_oauth_cls):
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_access_token.return_value = {
            "userinfo": {"sub": "alice", "groups": ["users"]},
        }

        client = self._make_client(restricted_groups=["admins"])
        with self.app.test_request_context():
            with self.assertRaisesRegex(
                OIDCAuthenticationError,
                r"alice.*restricted groups",
            ):
                client.authenticate()

    @patch("rfl.authentication.oidc.OAuth")
    def test_authenticate_oauth_error_translated(self, mock_oauth_cls):
        oauth_error = OAuthError(
            error="access_denied",
            description="User denied access",
        )
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_access_token.side_effect = oauth_error

        client = self._make_client()
        with self.app.test_request_context():
            with self.assertRaisesRegex(
                OIDCAuthenticationError,
                r"OIDC authorization failed: access_denied",
            ) as ctx:
                client.authenticate()

        self.assertIs(ctx.exception.__cause__, oauth_error)

    @patch("rfl.authentication.oidc.OAuth")
    def test_redirect_oauth_error_translated(self, mock_oauth_cls):
        oauth_error = OAuthError(error="mismatching_state")
        mock_client = MagicMock()
        mock_oauth_cls.return_value.register.return_value = mock_client
        mock_client.authorize_redirect.side_effect = oauth_error

        client = self._make_client()
        with self.app.test_request_context():
            with self.assertRaisesRegex(
                OIDCAuthenticationError,
                r"OIDC authorization redirect failed: mismatching_state",
            ) as ctx:
                client.redirect()

        self.assertIs(ctx.exception.__cause__, oauth_error)



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from pathlib import Path
import urllib
import ipaddress

import yaml

from .errors import SettingsDefinitionError


class SettingsDefinitionLoader:
    pass


class SettingsDefinitionLoaderYaml(SettingsDefinitionLoader):
    def __init__(self, raw: str = None, path: Path = None):
        try:
            if raw is not None:
                self.content = yaml.safe_load(raw)
                self.name = "definition:yaml:raw"
            elif path is not None:
                try:
                    with open(path) as fh:
                        self.content = yaml.safe_load(fh)
                except FileNotFoundError as err:
                    raise SettingsDefinitionError(
                        f"Settings definition file {path} not found"
                    ) from err
                self.name = f"def:yaml:{path}"
            else:
                raise SettingsDefinitionError(
                    "Either a raw string value or a path must be given to load YAML "
                    "settings definition"
                )
        except yaml.parser.ParserError as err:
            raise SettingsDefinitionError(
                f"Invalid YAML settings definition: {str(err)}"
            ) from err
        except yaml.scanner.ScannerError as err:
            raise SettingsDefinitionError(f"YAML scanner error: {str(err)}") from err


class SettingsParameterDefinition:
    # Map properties that can be found in definition file with corresponding
    # object attributes names.
    POSSIBLE_PROPERTIES = {
        "type": "_type",
        "content": "content",
        "default": "default",
        "choices": "choices",
        "doc": "doc",
        "ex": "example",
        "required": "required",
        "deprecated": "deprecated",
    }
    EXPECTED_TYPES = {
        "str": str,
        "password": str,
        "int": int,
        "float": float,
        "path": Path,
        "bool": bool,
        "uri": urllib.parse.ParseResult,
        "ip": ipaddress._BaseAddress,
        "network": ipaddress._BaseNetwork,
        "list": list,
    }

    def __init__(self, section: str, name: str, properties: dict):
        self.section = section
        self.name = name

        # Check for unsupported properties
        unsupported_properties = list(
            set(properties.keys()).difference(self.POSSIBLE_PROPERTIES.keys())
        )
        if len(unsupported_properties):
            raise SettingsDefinitionError(
                f"Unsupported properties found for {str(self)}: "
                f"{','.join(unsupported_properties)}"
            )

        for _property, attribute in self.POSSIBLE_PROPERTIES.items():
            value = None
            if _property in properties:
                value = properties[_property]
            setattr(self, attribute, value)

        # Check the type of defined parameter is actually supported
        if self._type not in self.EXPECTED_TYPES:
            raise SettingsDefinitionError(
                f"Unsupported type {self._type} for {str(self)}"
            )

        # Check content property is defined for list and check content type is actually
        # supported.
        if self._type == "list":
            if "content" not in properties:
                raise SettingsDefinitionError(
                    f"List content type for parameter {str(self)} must be defined"
                )
            elif self.content not in set(self.EXPECTED_TYPES.keys()) - {"list"}:
                raise SettingsDefinitionError(
                    f"Unsupported list content type {self.content} for {str(self)}"
                )
        # Check content property is not defined if not list
        if "content" in properties and self._type != "list":
            raise SettingsDefinitionError(
                f"Content property is forbidden for parameter {str(self)} with type "
                f"{self._type}"
            )

        # Check deprecated format
        if "deprecated" in properties:
            # check format
            if not isinstance(properties["deprecated"], dict):
                raise SettingsDefinitionError(
                    f"Unsupported format of {self} deprecated property"
                )
            if sorted(properties["deprecated"].keys()) != ["parameter", "section"]:
                raise SettingsDefinitionError(
                    f"Unexpected keys in {self} deprecated property"
                )
            if not isinstance(properties["deprecated"]["section"], str):
                raise SettingsDefinitionError(
                    f"Unexpected type of {self} deprecated section property"
                )
            if not isinstance(properties["deprecated"]["parameter"], str):
                raise SettingsDefinitionError(
                    f"Unexpected type of {self} deprecated parameter property"
                )
            # check no default value
            if self.default is not None:
                raise SettingsDefinitionError(
                    f"Deprecated parameter {self} cannot have default value"
                )
            # check not required
            if self.required:
                raise SettingsDefinitionError(
                    f"Deprecated parameter {self} cannot be required"
                )

        # If the default value is defined, convert to the expected type and validate it
        # is a valid choice.
        if self.default is not None:
            if self._type == "list":
                if not isinstance(self.default, list):
                    raise SettingsDefinitionError(
                        f"Default value {self.default} for parameter {str(self)} is "
                        "not a valid list"
                    )
                self.default = [
                    self._load_default(self.content, default)
                    for default in self.default
                ]
                for item in self.default:
                    self._validate_choice(item)
            else:
                self.default = self._load_default(self._type, self.default)
                self._validate_choice(self.default)

        if self.required is None:
            self.required = False
        elif not isinstance(self.required, bool):
            raise SettingsDefinitionError(
                f"Invalid boolean value of {str(self)} required property"
            )

    def _load_default(self, _type, default):
        """Verify default has the expected type and convert to advanded type if
        needed."""
        if _type == "path":
            return Path(default)
        elif _type == "uri":
            return urllib.parse.urlparse(default)
        elif _type == "ip":
            try:
                return ipaddress.ip_address(default)
            except ValueError as err:
                raise SettingsDefinitionError(
                    f"Invalid default ip address value for parameter {str(self)}"
                ) from err
        elif _type == "network":
            try:
                return ipaddress.ip_network(default)
            except ValueError as err:
                raise SettingsDefinitionError(
                    f"Invalid default ip network value for parameter {str(self)}"
                ) from err
        if not isinstance(default, self.EXPECTED_TYPES[_type]):
            raise SettingsDefinitionError(
                f"Default value {default} for parameter {str(self)} has not "
                f"the expected type {_type}"
            )
        return default

    def _validate_choice(self, default):
        """Check default is a valid choice or raise SettingsDefinitionError."""
        if self.choices is not None and default not in self.choices:
            raise SettingsDefinitionError(
                f"Default value {default} for parameter {str(self)} is not "
                f"one of possible choices {self.choices}"
            )

    def __str__(self):
        return f"[{self.section}]>{self.name}"


class SettingsSectionDefinition:
    SECTION_DOC_KEY = "_doc"

    def __init__(self, section: str, content: dict):
        self.name = section
        self.doc = content.get(self.SECTION_DOC_KEY)
        if self.doc is not None and not isinstance(self.doc, str):
            raise SettingsDefinitionError(
                f"Invalid type for section {self.SECTION_DOC_KEY} in [{section}]"
            )
        self.parameters = [
            SettingsParameterDefinition(section, parameter, properties)
            for parameter, properties in content.items()
            if parameter != self.SECTION_DOC_KEY
        ]

    def has_parameter(self, name: str) -> bool:
        for parameter in self.parameters:
            if parameter.name == name:
                return True
        return False

    def parameter(self, name: str):
        for parameter in self.parameters:
            if parameter.name == name:
                return parameter
        return None


class SettingsDefinition:
    def __init__(self, loader: SettingsDefinitionLoader):
        self.loader = loader
        self.sections = [
            SettingsSectionDefinition(section, content)
            for section, content in loader.content.items()
        ]
        self._validate_deprecated_references()

    def has_section(self, name: str) -> bool:
        for section in self.sections:
            if section.name == name:
                return True
        return False

    def section(self, name: str):
        for section in self.sections:
            if section.name == name:
                return section
        return None

    def _validate_deprecated_references(self):
        for section in self.sections:
            for parameter in section.parameters:
                if parameter.deprecated:
                    if not self.has_section(
                        parameter.deprecated["section"]
                    ) or not self.section(
                        parameter.deprecated["section"]
                    ).has_parameter(parameter.deprecated["parameter"]):
                        raise SettingsDefinitionError(
                            f"Invalid deprecated reference on parameter {parameter}"
                        )



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This tiny utility loads pyproject.toml file of current working directory project,
looking for this file in parents directory recursively up to filesystem root. If the
file is found, print the version of the project on stdout. Else, print an error.
"""

import sys
from pathlib import Path

# First try to import Python standard toml library (available in Python 3.11+) and
# fallback to external library.
try:
    import tomllib
except ImportError:
    import tomli as tomllib

from .errors import BuildError


def project_version():
    current_dir = Path.cwd()
    found = False
    while True:
        pyproject_path = current_dir.joinpath("pyproject.toml")
        if pyproject_path.exists():
            found = True
            break
        else:
            # Stop iteration when filesystem root is reached.
            if str(current_dir) == current_dir.root:
                break
            # Jump to parent directory.
            current_dir = current_dir.parent

    if not found:
        raise BuildError(
            "Unable to find project pyproject.toml file of current working directory",
        )
    content = tomllib.load(open(pyproject_path, "rb"))
    try:
        return content["project"]["version"]
    except KeyError as err:
        raise BuildError(
            f"Unable to extract version from project file {pyproject_path}"
        ) from err


def main():
    try:
        print(project_version())
    except BuildError as err:
        print(str(err), file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright (c) 2024 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest

from rfl.core.utils import shlex_join


class TestUtils(unittest.TestCase):
    def test_shlex_join(self):
        self.assertEqual(shlex_join(["cmd", "arg1", "arg2"]), "cmd arg1 arg2")
        self.assertEqual(
            shlex_join(["cmd", "arg1", "arg2 with space"]), "cmd arg1 'arg2 with space'"
        )



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
from unittest.mock import patch, Mock
from pathlib import Path
import urllib
import ssl

import ldap
import ldap.filter

from rfl.authentication.ldap import LDAPAuthentifier
from rfl.authentication.errors import LDAPAuthenticationError


class MockLDAPObject:
    pass


class TestLDAPAuthentifier(unittest.TestCase):
    def setUp(self):
        self.authentifier = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
        )

    def test_connection(self):
        # With classic ldap uri (no SSL/TLS)
        self.authentifier.connection()

    @patch("rfl.authentication.ldap.ldap")
    def test_connection_ssl_cert(self, mock_ldap):
        mock_ldap_object = mock_ldap.initialize.return_value

        # With ldap URI, check set_option is not called
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_not_called()

        # With ldaps URI and no CA certificate path, check LDAP server certificate is
        # required and validated with default system OpenSSL certificates directory.
        self.authentifier.uri = urllib.parse.urlparse("ldaps://localhost")
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_REQUIRE_CERT, mock_ldap.OPT_X_TLS_DEMAND
        )
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_CACERTDIR, ssl.get_default_verify_paths().openssl_capath
        )
        mock_ldap_object.reset_mock()

        # With CA certificate path, check LDAP server certificate is required and
        # validated with provided CA certificate.
        cert = Path("/dev/null")
        self.authentifier.cacert = cert
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_REQUIRE_CERT, mock_ldap.OPT_X_TLS_DEMAND
        )
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_CACERTFILE, str(cert)
        )
        mock_ldap_object.reset_mock()

    @patch("rfl.authentication.ldap.ldap")
    def test_connection_starttls(self, mock_ldap):
        mock_ldap_object = mock_ldap.initialize.return_value

        # With ldaps URI (and starttls set to False by default), check LDAP server
        # certificate is required and start_tls_s is NOT called.
        self.authentifier.uri = urllib.parse.urlparse("ldaps://localhost")
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_REQUIRE_CERT, mock_ldap.OPT_X_TLS_DEMAND
        )
        mock_ldap_object.start_tls_s.assert_not_called()
        mock_ldap_object.reset_mock()

        # With ldap URI and starttls set to True, check LDAP server certificate is
        # required and start_tls_s is called.
        self.authentifier.uri = urllib.parse.urlparse("ldap://localhost")
        self.authentifier.starttls = True
        self.authentifier.connection()
        mock_ldap_object.set_option.assert_any_call(
            mock_ldap.OPT_X_TLS_REQUIRE_CERT, mock_ldap.OPT_X_TLS_DEMAND
        )
        mock_ldap_object.start_tls_s.assert_called_once()

    @patch.object(ldap.ldapobject.LDAPObject, "start_tls_s")
    def test_connection_tls_errors(self, mock_start_tls_s):
        self.authentifier.uri = urllib.parse.urlparse("ldap://localhost")
        self.authentifier.starttls = True
        mock_start_tls_s.side_effect = ldap.CONNECT_ERROR("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to connect to LDAP server with STARTTLS: fail$",
        ):
            self.authentifier.connection()
        mock_start_tls_s.side_effect = ldap.SERVER_DOWN("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^LDAP server {self.authentifier.uri.geturl()} is unreachable$",
        ):
            self.authentifier.connection()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_disabled(self, mock_ldap_initialize):
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value

        self.assertEqual(
            self.authentifier._lookup_user_dn("john"),
            f"{self.authentifier.user_name_attribute}=john,"
            f"{self.authentifier.user_base}",
        )
        mock_ldap_object.simple_bind_s.assert_not_called()
        mock_ldap_object.search_s.assert_not_called()
        mock_ldap_object.unbind_s.assert_not_called()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled(self, mock_ldap_initialize):
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.return_value = [
            (
                f"uid=john,ou=admins,{self.authentifier.user_base}",
                {"cn": [b"John Doe"]},
            )
        ]

        self.assertEqual(
            self.authentifier._lookup_user_dn("john"),
            f"uid=john,ou=admins,{self.authentifier.user_base}",
        )
        mock_ldap_object.simple_bind_s.assert_not_called()
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_bind_dn(self, mock_ldap_initialize):
        # define bind dn/password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        self.authentifier.bind_password = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.return_value = [
            (
                f"uid=john,ou=admins,{self.authentifier.user_base}",
                {"cn": [b"John Doe"]},
            )
        ]

        self.assertEqual(
            self.authentifier._lookup_user_dn("john"),
            f"uid=john,ou=admins,{self.authentifier.user_base}",
        )
        mock_ldap_object.simple_bind_s.assert_called_once_with(
            self.authentifier.bind_dn, self.authentifier.bind_password
        )
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_escape_special_chars(self, mock_ldap_initialize):
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        user_with_special_chars = "John Doe(user)"
        escaped_user = ldap.filter.escape_filter_chars(user_with_special_chars)
        mock_ldap_object.search_s.return_value = [
            (
                f"uid={user_with_special_chars},ou=admins,{self.authentifier.user_base}",
                {"cn": [b"John Doe(user)"]},
            )
        ]

        self.assertEqual(
            self.authentifier._lookup_user_dn(user_with_special_chars),
            f"uid={user_with_special_chars},ou=admins,{self.authentifier.user_base}",
        )
        # Verify the filter contains escaped user value
        call_args = mock_ldap_object.search_s.call_args
        search_filter = call_args[0][2]
        # Verify the filter contains escaped user value
        self.assertIn(
            f"({self.authentifier.user_name_attribute}={escaped_user})", search_filter
        )
        # The unescaped value should NOT be in the filter
        self.assertNotIn(
            f"({self.authentifier.user_name_attribute}={user_with_special_chars})",
            search_filter,
        )
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_bind_dn_missing_password(
        self, mock_ldap_initialize
    ):
        # define bind_dn without password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value

        # Check exception is raised when bind_dn is set without bind_password
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Password to authenticate with bind DN uid=read,ou=apps,dc=corp,dc=org "
            r"is required$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_not_called()
        mock_ldap_object.unbind_s.assert_not_called()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_bind_dn_invalid_credentials(
        self, mock_ldap_initialize
    ):
        # define bind dn/password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        self.authentifier.bind_password = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.simple_bind_s.side_effect = ldap.INVALID_CREDENTIALS("fail")

        # Check exception is raised with LDAP fails due to invalid credential
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Invalid bind DN or password$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_not_called()
        mock_ldap_object.unbind_s.assert_not_called()

    @patch.object(ldap, "initialize")
    def test_lookup_user_ldap_server_down_error(self, mock_ldap_initialize):
        # define bind dn/password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        self.authentifier.bind_password = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.simple_bind_s.side_effect = ldap.SERVER_DOWN("fail")

        # Check exception is raised due to LDAP server down
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^LDAP server {self.authentifier.uri.geturl()} is unreachable$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_not_called()
        mock_ldap_object.unbind_s.assert_not_called()

    @patch.object(ldap, "initialize")
    def test_lookup_user_ldap_operations_error(self, mock_ldap_initialize):
        # define bind dn/password
        self.authentifier.bind_dn = "uid=read,ou=apps,dc=corp,dc=org"
        self.authentifier.bind_password = "uid=read,ou=apps,dc=corp,dc=org"
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.side_effect = ldap.OPERATIONS_ERROR("fail")

        # Check exception is raised due to LDAP server down
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Operations error on user DN lookup: fail$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_not_found(self, mock_ldap_initialize):
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.return_value = []

        # Check exception is raised due to no result found
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to find user john in base ou=people,dc=corp,dc=org$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(ldap, "initialize")
    def test_lookup_user_dn_enabled_too_many_results(self, mock_ldap_initialize):
        # enable user DN lookup
        self.authentifier.lookup_user_dn = True
        # setup LDAP mock
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.search_s.return_value = [
            (
                f"uid=john,ou=admins,{self.authentifier.user_base}",
                {"cn": [b"John Doe"]},
            ),
            (
                f"uid=alice,ou=admins,{self.authentifier.user_base}",
                {"cn": [b"Alice Doe"]},
            ),
        ]
        # Check exception is raised due to too many results found
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Too many users found \(2\) with username john in base "
            r"ou=people,dc=corp,dc=org$",
        ):
            self.authentifier._lookup_user_dn("john")
        mock_ldap_object.search_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch("rfl.authentication.ldap.ldap")
    def test_login_ok(
        self, mock_ldap, mock_get_groups, mock_get_user_info, mock_lookup_user_dn
    ):
        # setup mocks return values
        mock_get_groups.return_value = ["group1", "group2"]
        mock_get_user_info.return_value = ("John Doe", 42)
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"
        mock_ldap_object = mock_ldap.initialize.return_value

        # call method
        user = self.authentifier.login("john", "SECR3T")

        # verify mock calls
        mock_ldap_object.unbind_s.assert_called_once()
        mock_ldap_object.simple_bind_s.assert_called_once_with(
            "uid=john,ou=people,dc=corp,dc=org", "SECR3T"
        )
        mock_get_user_info.assert_called_once_with(
            mock_ldap_object, "uid=john,ou=people,dc=corp,dc=org"
        )
        mock_get_groups.assert_called_once_with(
            mock_ldap_object, "john", "uid=john,ou=people,dc=corp,dc=org", 42
        )

        # verify return value
        self.assertEqual(user.login, "john")
        self.assertEqual(user.fullname, "John Doe")
        self.assertEqual(user.groups, ["group1", "group2"])

    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch("rfl.authentication.ldap.ldap")
    def test_login_not_in_restricted_group(
        self, mock_ldap, mock_get_groups, mock_get_user_info, mock_lookup_user_dn
    ):
        # setup mocks return values
        self.authentifier.restricted_groups = ["group3", "group4"]
        mock_get_groups.return_value = ["group1", "group2"]
        mock_get_user_info.return_value = ("John Doe", 42)
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"

        # call method
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^User john is not member of restricted groups$",
        ):
            self.authentifier.login("john", "SECR3T")

    def test_login_missing_user_or_password(self):
        with self.assertRaisesRegex(
            LDAPAuthenticationError, "Invalid authentication request"
        ):
            self.authentifier.login("john", None)
            self.authentifier.login(None, "SECR3T")

    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(ldap.ldapobject.LDAPObject, "simple_bind_s")
    def test_login_errors(self, mock_simple_bind_s, mock_lookup_user_dn):
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"
        mock_simple_bind_s.side_effect = ldap.SERVER_DOWN("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^LDAP server {self.authentifier.uri.geturl()} is unreachable$",
        ):
            self.authentifier.login("john", "SECR3T")
        mock_simple_bind_s.side_effect = ldap.INVALID_CREDENTIALS("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError, "^Invalid user or password$"
        ):
            self.authentifier.login("john", "SECR3return_valueT")
        mock_simple_bind_s.side_effect = ldap.UNWILLING_TO_PERFORM("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError, "^LDAP server is unwilling to perform: fail$"
        ):
            self.authentifier.login("john", "SECR3T")

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(LDAPAuthentifier, "_bind")
    @patch.object(LDAPAuthentifier, "connection")
    def test_login_no_bind_lookup_as_user_true(
        self,
        mock_connection,
        mock_bind,
        mock_lookup_user_dn,
        mock_get_user_info,
        mock_get_groups,
    ):
        # setup mocks return values
        mock_get_groups.return_value = ["group1", "group2"]
        mock_get_user_info.return_value = ("John Doe", 42)
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"
        mock_connection.return_value.simple_bind_s.return_value = None

        # if lookup_as_user is False, login() must not call _bind().
        self.authentifier.lookup_as_user = True
        self.authentifier.login("john", "SECR3T")
        mock_bind.assert_not_called()
        mock_connection.return_value.simple_bind_s.assert_called_once_with(
            "uid=john,ou=people,dc=corp,dc=org", "SECR3T"
        )

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch.object(LDAPAuthentifier, "_lookup_user_dn")
    @patch.object(LDAPAuthentifier, "_bind")
    @patch.object(LDAPAuthentifier, "connection")
    def test_login_single_bind_lookup_as_user_false(
        self,
        mock_connection,
        mock_bind,
        mock_lookup_user_dn,
        mock_get_user_info,
        mock_get_groups,
    ):
        # setup mocks return values
        mock_get_groups.return_value = ["group1", "group2"]
        mock_get_user_info.return_value = ("John Doe", 42)
        mock_lookup_user_dn.return_value = "uid=john,ou=people,dc=corp,dc=org"

        # if lookup_as_user is False, login() should call _bind() once.
        self.authentifier.lookup_as_user = False
        self.authentifier.login("john", "SECR3T")
        mock_bind.assert_called_once()
        mock_connection.return_value.simple_bind_s.assert_called_once_with(
            "uid=john,ou=people,dc=corp,dc=org", "SECR3T"
        )

    def test_user_info(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            (
                "uid=john,ou=people,dc=corp,dc=org",
                {"cn": [b"John Doe"], "gidNumber": [b"42"]},
            )
        ]
        fullname, gid = self.authentifier._get_user_info(
            connection, "uid=john,ou=people,dc=corp,dc=org"
        )
        self.assertEqual(fullname, "John Doe")
        self.assertEqual(gid, 42)

    def test_user_info_fullname_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # If the user entries in LDAP directory do not contain attributes whose name
        # matches user_fullname_attribute or user_primary_group_attribute, search_s
        # returns a dict with missing keys in the second element of the result tuple.
        #
        # Test missing user_fullname_attribute
        connection.search_s.return_value = [
            (
                "uid=john,ou=people,dc=corp,dc=org",
                {"gidNumber": [b"42"]},
            )
        ]
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to extract user full name with "
            rf"{self.authentifier.user_fullname_attribute} attribute from user "
            r"entries$",
        ):
            self.authentifier._get_user_info(
                connection, "uid=john,ou=people,dc=corp,dc=org"
            )

    def test_user_info_primary_group_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # Test missing gidNumber
        connection.search_s.return_value = [
            (
                "uid=john,ou=people,dc=corp,dc=org",
                {"cn": [b"John Doe"]},
            )
        ]
        with self.assertLogs("rfl", level="WARNING") as log:
            fullname, gid = self.authentifier._get_user_info(
                connection, "uid=john,ou=people,dc=corp,dc=org"
            )
            self.assertEqual(
                [
                    "WARNING:rfl.authentication.ldap:Unable to extract user primary "
                    "group with gidNumber attribute from user entry"
                ],
                log.output,
            )
        self.assertEqual(fullname, "John Doe")
        self.assertIsNone(gid)

    def test_user_info_class_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # If entries with class user_class is not found in user_base subtree,
        # search_s returns an empty list.
        connection.search_s.return_value = []
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^User not found in LDAP with class {self.authentifier.user_class}$",
        ):
            self.authentifier._get_user_info(
                connection, "uid=john,ou=people,dc=corp,dc=org"
            )

    def test_user_info_dn_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # When user DN is not found in LDAP, ldap module raises NO_SUCH_OBJECT
        # exception.
        connection.search_s.side_effect = ldap.NO_SUCH_OBJECT("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^Unable to find user DN uid=john,{self.authentifier.user_base}$",
        ):
            self.authentifier._get_user_info(
                connection, "uid=john,ou=people,dc=corp,dc=org"
            )

    def test_custom_primary_attribute(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        self.authentifier.user_primary_group_attribute = "primaryGroupId"
        dn = "uid=jane,ou=people,dc=corp,dc=org"
        connection.search_s.return_value = [
            (
                dn,
                {
                    "cn": [b"Jane Smith"],
                    self.authentifier.user_primary_group_attribute: [b"42"],
                },
            )
        ]
        fullname, gid = self.authentifier._get_user_info(connection, dn)
        connection.search_s.assert_called_once_with(
            dn,
            ldap.SCOPE_BASE,
            f"(objectClass={self.authentifier.user_class})",
            [
                self.authentifier.user_fullname_attribute,
                self.authentifier.user_primary_group_attribute,
            ],
        )
        self.assertEqual(fullname, "Jane Smith")
        self.assertEqual(gid, 42)

    def test_get_groups(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {"cn": [b"scientists"]}),
            ("cn=biology,ou=groups,dc=corp,dc=org", {"cn": [b"biology"]}),
        ]
        groups = self.authentifier._get_groups(
            connection, "john", "uid=john,ou=people,dc=corp,dc=org", 42
        )
        self.assertEqual(groups, ["scientists", "biology"])

    def test_get_groups_escape_special_chars(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {"cn": [b"scientists"]}),
        ]
        user_name_with_special = "John Doe(user)"
        user_dn_with_special = "uid=John Doe(user),ou=people,dc=corp,dc=org"
        escaped_user_name = ldap.filter.escape_filter_chars(user_name_with_special)
        escaped_user_dn = ldap.filter.escape_filter_chars(user_dn_with_special)
        gid = 42
        groups = self.authentifier._get_groups(
            connection, user_name_with_special, user_dn_with_special, gid
        )
        self.assertEqual(groups, ["scientists"])

        connection.search_s.assert_called_once_with(
            self.authentifier.group_base,
            ldap.SCOPE_SUBTREE,
            f"(&(|(objectClass=posixGroup)(objectClass=groupOfNames))"
            f"(|(memberUid={escaped_user_name})"
            f"(member={escaped_user_dn})(gidNumber={gid})))",
            [self.authentifier.group_name_attribute],
        )

    def test_get_groups_without_gid(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {"cn": [b"scientists"]}),
            ("cn=biology,ou=groups,dc=corp,dc=org", {"cn": [b"biology"]}),
        ]
        user = "john"
        dn = "uid=john,ou=people,dc=corp,dc=org"
        gid = 42
        # First call with gid and check LDAP search filter.
        groups = self.authentifier._get_groups(connection, user, dn, gid)
        self.assertEqual(groups, ["scientists", "biology"])
        connection.search_s.assert_called_once_with(
            self.authentifier.group_base,
            ldap.SCOPE_SUBTREE,
            "(&(|(objectClass=posixGroup)(objectClass=groupOfNames))"
            f"(|(memberUid={user})(member={dn})(gidNumber={gid})))",
            [self.authentifier.group_name_attribute],
        )
        connection.search_s.reset_mock()
        # Then a second call with undefined gid must remove gidNumber from LDAP search
        # filter.
        groups = self.authentifier._get_groups(connection, user, dn, None)
        self.assertEqual(groups, ["scientists", "biology"])
        connection.search_s.assert_called_once_with(
            self.authentifier.group_base,
            ldap.SCOPE_SUBTREE,
            "(&(|(objectClass=posixGroup)(objectClass=groupOfNames))"
            f"(|(memberUid={user})(member={dn})))",
            [self.authentifier.group_name_attribute],
        )

    def test_groups_base_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # When group base DN is not found in LDAP, ldap module raises NO_SUCH_OBJECT
        # exception.
        connection.search_s.side_effect = ldap.NO_SUCH_OBJECT("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^Unable to find group base {self.authentifier.group_base}$",
        ):
            self.authentifier._get_groups(
                connection, "john", "uid=john,ou=people,dc=corp,dc=org", 42
            )

    def test_get_groups_name_attribute_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        self.authentifier.group_name_attribute = "fail"
        # If the group entries in LDAP directory does not contain attributes whose name
        # matches group_name_attribute, search_s returns an empty dict in the second
        # element of the result tuple.
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {}),
            ("cn=biology,ou=groups,dc=corp,dc=org", {}),
        ]
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to extract group name with fail attribute from group entries$",
        ):
            self.authentifier._get_groups(
                connection, "john", "uid=john,ou=people,dc=corp,dc=org", 42
            )

    def test_get_groups_class_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # If entries with one of group_object_classes is not found in group_base
        # subtree, search_s returns an empty list.
        connection.search_s.return_value = []
        with self.assertLogs("rfl.authentication.ldap", level="WARNING") as cm:
            groups = self.authentifier._get_groups(
                connection, "john", "uid=john,ou=people,dc=corp,dc=org", 42
            )
        self.assertEqual(groups, [])
        self.assertEqual(
            cm.output,
            [
                "WARNING:rfl.authentication.ldap:Unable to find groups in LDAP for user"
                " john or gidNumber 42"
            ],
        )
        # Test log message without gid
        with self.assertLogs("rfl.authentication.ldap", level="WARNING") as cm:
            groups = self.authentifier._get_groups(
                connection, "john", "uid=john,ou=people,dc=corp,dc=org", None
            )
        self.assertEqual(groups, [])
        self.assertEqual(
            cm.output,
            [
                "WARNING:rfl.authentication.ldap:Unable to find groups in LDAP for user"
                " john"
            ],
        )

    def test_custom_group_object_classes(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("cn=scientists,ou=groups,dc=corp,dc=org", {"cn": [b"scientists"]}),
            ("cn=biology,ou=groups,dc=corp,dc=org", {"cn": [b"biology"]}),
        ]
        group_object_class = "group"
        login = "john"
        gid = 42
        self.authentifier.group_object_classes = [group_object_class]
        groups = self.authentifier._get_groups(
            connection, login, f"uid={login},ou=people,dc=corp,dc=org", gid
        )
        connection.search_s.assert_called_once_with(
            self.authentifier.group_base,
            ldap.SCOPE_SUBTREE,
            f"(&(|(objectClass={group_object_class}))(|(memberUid={login})"
            f"(member=uid={login},ou=people,dc=corp,dc=org)(gidNumber={gid})))",
            [self.authentifier.group_name_attribute],
        )
        self.assertEqual(groups, ["scientists", "biology"])

    def test_in_restricted_groups(self):
        # By default, restricted groups are unset, _in_restricted_groups must return
        # True in all cases.
        self.assertTrue(self.authentifier._in_restricted_groups([]))
        self.assertTrue(self.authentifier._in_restricted_groups(["users", "admins"]))

        # If restricted groups are set, _in_restricted_groups must return if at least
        # one group in argument matches one restricted group.

        # Test with one restricted group
        self.authentifier.restricted_groups = ["admins"]
        self.assertFalse(self.authentifier._in_restricted_groups([]))
        self.assertFalse(
            self.authentifier._in_restricted_groups(["users", "scientists"])
        )
        self.assertTrue(self.authentifier._in_restricted_groups(["admins"]))
        self.assertTrue(
            self.authentifier._in_restricted_groups(["users", "scientists", "admins"])
        )

        # Test with multiple restricted groups
        self.authentifier.restricted_groups = ["admins", "scientists"]
        self.assertFalse(self.authentifier._in_restricted_groups([]))
        self.assertFalse(self.authentifier._in_restricted_groups(["users"]))
        self.assertTrue(
            self.authentifier._in_restricted_groups(["users", "scientists"])
        )
        self.assertTrue(self.authentifier._in_restricted_groups(["admins"]))
        self.assertTrue(
            self.authentifier._in_restricted_groups(["users", "scientists", "admins"])
        )

    def test_list_user_dn(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = [
            ("uid=john,ou=people,dc=corp,dc=org", {"uid": [b"john"]}),
            ("uid=marie,ou=people,dc=corp,dc=org", {"uid": [b"marie"]}),
        ]
        results = self.authentifier._list_user_dn(connection)
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0], ("john", "uid=john,ou=people,dc=corp,dc=org"))
        self.assertEqual(results[1], ("marie", "uid=marie,ou=people,dc=corp,dc=org"))

    def test_list_user_dn_not_found(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # When user DN is not found in LDAP, ldap module raises NO_SUCH_OBJECT
        # exception.
        connection.search_s.side_effect = ldap.NO_SUCH_OBJECT("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to find user base ou=people,dc=corp,dc=org$",
        ):
            self.authentifier._list_user_dn(connection)

    def test_list_user_dn_no_result(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        connection.search_s.return_value = []
        with self.assertLogs("rfl", level="WARNING") as lc:
            results = self.authentifier._list_user_dn(connection)
            self.assertEqual(
                [
                    "WARNING:rfl.authentication.ldap:Unable to find users in LDAP in "
                    "base ou=people,dc=corp,dc=org subtree"
                ],
                lc.output,
            )
        self.assertEqual(len(results), 0)

    def test_list_user_dn_no_user_name_attribute(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # All results miss the user name attribute, _list_user_dn() is expected to raise
        # LDAPAuthenticationError.
        connection.search_s.return_value = [
            ("uid=john,ou=people,dc=corp,dc=org", {}),
            ("uid=jane,ou=people,dc=corp,dc=org", {}),
        ]
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Unable to extract user uid from user entries$",
        ):
            self.authentifier._list_user_dn(connection)

    def test_list_user_dn_missing_user_name_attribute(self):
        connection = Mock(spec=ldap.ldapobject.LDAPObject)
        # At least one result has the user name attribute, _list_user_dn() must return
        # these results and log warning message for other dn that miss this attribute.
        connection.search_s.return_value = [
            ("uid=john,ou=people,dc=corp,dc=org", {}),
            ("uid=jane,ou=people,dc=corp,dc=org", {"uid": [b"jane"]}),
        ]
        with self.assertLogs("rfl", level="INFO") as lc:
            results = self.authentifier._list_user_dn(connection)

        self.assertEqual(len(results), 1)
        self.assertEqual(results[0], ("jane", "uid=jane,ou=people,dc=corp,dc=org"))

        # Check warning message to indicate user name attribute has not been found in
        # john user entry has been sent
        self.assertEqual(
            [
                "WARNING:rfl.authentication.ldap:Unable to find uid from user entry "
                "uid=john,ou=people,dc=corp,dc=org"
            ],
            lc.output,
        )

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_list_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch("rfl.authentication.ldap.ldap")
    def test_users(
        self, mock_ldap, mock_get_user_info, mock_list_user_dn, mock_get_groups
    ):
        # Setup mocks return values
        mock_list_user_dn.return_value = [
            ("john", "uid=john,ou=people,dc=corp,dc=org"),
            ("marie", "uid=marie,ou=people,dc=corp,dc=org"),
        ]
        mock_get_user_info.side_effect = [("John Doe", 42), ("Marie Magic", 43)]
        mock_ldap_object = mock_ldap.initialize.return_value

        # Call method (without groups)
        users = self.authentifier.users()

        # Verify mock calls
        mock_ldap_object.unbind_s.assert_called_once()
        mock_list_user_dn.assert_called_once_with(mock_ldap_object)
        mock_get_user_info.assert_called()

        # Verify return value
        self.assertEqual(len(users), 2)
        self.assertEqual(users[0].login, "john")
        self.assertEqual(users[0].fullname, "John Doe")
        self.assertEqual(users[0].groups, [])
        self.assertEqual(users[1].login, "marie")
        self.assertEqual(users[1].fullname, "Marie Magic")
        self.assertEqual(users[1].groups, [])

        # Reset mock and re-inject side effects
        mock_ldap_object.reset_mock()
        mock_list_user_dn.reset_mock()
        mock_get_user_info.reset_mock()
        mock_get_user_info.side_effect = [("John Magic", 45), ("Marie Doe", 46)]
        mock_get_groups.side_effect = [["admin", "users"], ["biology", "users"]]

        # Call method (with groups)
        users = self.authentifier.users(True)

        # Verify mock calls
        mock_ldap_object.unbind_s.assert_called_once()
        mock_list_user_dn.assert_called_once_with(mock_ldap_object)
        mock_get_user_info.assert_called()

        # Verify return value
        self.assertEqual(len(users), 2)
        self.assertEqual(users[0].login, "john")
        self.assertEqual(users[0].fullname, "John Magic")
        self.assertEqual(users[0].groups, ["admin", "users"])
        self.assertEqual(users[1].login, "marie")
        self.assertEqual(users[1].fullname, "Marie Doe")
        self.assertEqual(users[1].groups, ["biology", "users"])

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_list_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch("rfl.authentication.ldap.ldap")
    def test_users_restricted_groups(
        self, mock_ldap, mock_get_user_info, mock_list_user_dn, mock_get_groups
    ):
        self.authentifier.restricted_groups = ["biology"]

        # Setup mocks return values
        mock_list_user_dn.return_value = [
            ("john", "uid=john,ou=people,dc=corp,dc=org"),
            ("marie", "uid=marie,ou=people,dc=corp,dc=org"),
        ]
        mock_get_user_info.side_effect = [("John Magic", 45), ("Marie Doe", 46)]
        mock_get_groups.side_effect = [["admin", "users"], ["biology", "users"]]

        # Call users method with groups retrieval
        with self.assertLogs("rfl", level="DEBUG") as lc:
            users = self.authentifier.users(True)

        # Verify return value. Only user Marie in restricted biology group must be
        # present.
        self.assertEqual(len(users), 1)
        self.assertEqual(users[0].login, "marie")
        self.assertEqual(users[0].fullname, "Marie Doe")
        self.assertEqual(users[0].groups, ["biology", "users"])

        # Check debug message to indicate john is discarded has been sent
        self.assertEqual(
            [
                "DEBUG:rfl.authentication.ldap:Discarding user john not member of "
                "restricted groups"
            ],
            lc.output,
        )

    @patch.object(LDAPAuthentifier, "_get_groups")
    @patch.object(LDAPAuthentifier, "_list_user_dn")
    @patch.object(LDAPAuthentifier, "_get_user_info")
    @patch("rfl.authentication.ldap.ldap")
    def test_users_bind_dn(
        self, mock_ldap, mock_get_user_info, mock_list_user_dn, mock_get_groups
    ):
        self.authentifier.bind_dn = "uid=hey,ou=people,dc=corp,dc=org"
        self.authentifier.bind_password = "secr3t"

        mock_ldap_object = mock_ldap.initialize.return_value

        # Call method
        self.authentifier.users()

        # Check simple_bind_s is called when bind_dn and bind_password are defined
        mock_ldap_object.simple_bind_s.assert_called_once()
        mock_ldap_object.unbind_s.assert_called_once()

    @patch("rfl.authentication.ldap.ldap")
    def test_users_bind_dn_missing_password(self, mock_ldap):
        self.authentifier.bind_dn = "uid=hey,ou=people,dc=corp,dc=org"

        # Check exception is raised when bind_dn is set without bind_password
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Password to authenticate with bind DN uid=hey,ou=people,dc=corp,dc=org "
            r"is required$",
        ):
            self.authentifier.users()

    @patch.object(ldap, "initialize")
    def test_users_bind_dn_invalid_credentials(self, mock_ldap_initialize):
        mock_ldap_object = mock_ldap_initialize.return_value
        mock_ldap_object.simple_bind_s.side_effect = ldap.INVALID_CREDENTIALS("fail")

        self.authentifier.bind_dn = "uid=hey,ou=people,dc=corp,dc=org"
        self.authentifier.bind_password = "secr3t"

        # Check exception is raised with LDAP fails due to invalid credential
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Invalid bind DN or password$",
        ):
            self.authentifier.users()

    @patch.object(ldap.ldapobject.LDAPObject, "search_s")
    def test_users_ldap_server_down_error(self, mock_search_s):
        mock_search_s.side_effect = ldap.SERVER_DOWN("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            rf"^LDAP server {self.authentifier.uri.geturl()} is unreachable$",
        ):
            self.authentifier.users()

    @patch.object(ldap.ldapobject.LDAPObject, "search_s")
    def test_users_ldap_operations_error(self, mock_search_s):
        mock_search_s.side_effect = ldap.OPERATIONS_ERROR("fail")
        with self.assertRaisesRegex(
            LDAPAuthenticationError,
            r"^Operations error on users search: fail$",
        ):
            self.authentifier.users()


class TestLDAPAuthentifierInit(unittest.TestCase):
    def test_lookup_as_user_auto_bind_dn(self):
        # lookup_as_user is None, bind_dn and bind_password are set, should do lookup
        # with service credentials.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            bind_dn="uid=read,ou=apps,dc=corp,dc=org",
            bind_password="SECR3T",
            lookup_as_user=None,
        )
        self.assertFalse(auth.lookup_as_user)

    def test_lookup_as_user_auto_no_bind_user(self):
        # lookup_as_user is None, bind_dn and bind_password are not set, should do
        # lookup as user.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            bind_dn=None,
            bind_password=None,
            lookup_as_user=None,
        )
        self.assertTrue(auth.lookup_as_user)

    def test_lookup_as_user_auto_no_bind_dn(self):
        # lookup_as_user is None, bind_dn is not set, should do lookup as user.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            bind_dn=None,
            bind_password="SECR3T",
            lookup_as_user=None,
        )
        self.assertTrue(auth.lookup_as_user)

    def test_lookup_as_user_auto_no_bind_password(self):
        # lookup_as_user is None, bind_password is not set, should do lookup as user.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            bind_dn="uid=read,ou=apps,dc=corp,dc=org",
            bind_password=None,
            lookup_as_user=None,
        )
        self.assertTrue(auth.lookup_as_user)

    def test_lookup_as_user_enabled(self):
        # lookup_as_user is True, should do lookup as user.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            lookup_as_user=True,
        )
        self.assertTrue(auth.lookup_as_user)

    def test_lookup_as_user_disabled(self):
        # lookup_as_user is False, should do lookup with service credentials.
        auth = LDAPAuthentifier(
            uri=urllib.parse.urlparse("ldap://localhost"),
            user_base="ou=people,dc=corp,dc=org",
            group_base="ou=groups,dc=corp,dc=org",
            lookup_as_user=False,
        )
        self.assertFalse(auth.lookup_as_user)



# Copyright (c) 2024 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

"""This module contains a backport of asyncio.run() compatible with Python 3.6. It is
designed to be called through asyncio_run() function which detects Python interpreter
version and call appropriate implementation consequently."""

import asyncio
import sys


def _cancel_all_tasks(loop):
    """Backport of standard Python asyncio.runners._cancel_all_tasks() for
    Python 3.6."""
    to_cancel = asyncio.Task.all_tasks(loop)
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(
        asyncio.tasks.gather(*to_cancel, loop=loop, return_exceptions=True)
    )

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )


def _asyncio_run_backport(main):
    """Backport of standard Python asyncio.run() for Python 3.6."""
    if asyncio.events._get_running_loop() is not None:
        raise RuntimeError("asyncio.run() cannot be called from a running event loop")

    if not asyncio.coroutines.iscoroutine(main):
        raise ValueError("a coroutine was expected, got {!r}".format(main))

    loop = asyncio.events.new_event_loop()
    try:
        asyncio.events.set_event_loop(loop)
        return loop.run_until_complete(main)
    finally:
        try:
            _cancel_all_tasks(loop)
            loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            asyncio.events.set_event_loop(None)
            loop.close()


def asyncio_run(main):
    """Wrapper of asyncio.run() or asyncio_run_backport when running on Python 3.6."""
    if sys.version_info[0] >= 3 and sys.version_info[1] >= 7:
        return asyncio.run(main)
    else:
        return _asyncio_run_backport(main)



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
from pathlib import Path
from io import StringIO
from unittest.mock import patch
import ipaddress

from rfl.settings.definition import SettingsDefinition, SettingsDefinitionLoaderYaml
from rfl.settings.loaders import RuntimeSettingsSiteLoaderIni
from rfl.settings import RuntimeSettings, RuntimeSettingsSection

from rfl.settings.errors import SettingsDefinitionError, SettingsOverrideError

VALID_DEFINITION = """
---
section1:
  param_str:
    type: str
    default: value1

section2:
    param_int:
        type: int
        default: 10
        doc: documentation of param2
        ex: 500
        choices:
        - 10
        - 100
        - 500
    param_path:
        type: path
        default: /path/to/file
        doc: documentation of param3
        ex: /other/path/to/file
    param_uri:
        type: uri
        default: https://localhost:5900/resources
    param_float:
        type: float
    param_bool:
        type: bool
        default: yes
    param_list:
        type: list
        content: str
        default:
        - value1
        - value2
    param_required:
        type: str
        required: true
    param_password:
        type: password
    param_ip:
        type: ip
    param_network:
        type: network
    param_list_uri:
        type: list
        content: uri
    param_list_ip:
        type: list
        content: ip
    param_list_network:
        type: list
        content: network
    param_deprecated:
        type: str
        deprecated:
            section: section1
            parameter: param_str
"""

VALID_SITE = """
[section1]
param_str = site_value1

[section2]
param_path = /site/path/to/file
param_list =
  value3
  #value4
  value5
param_required = required_value
param_password = SECR3T
param_ip = 127.0.0.1
param_network = 127.0.0.0/24
param_list_uri =
  http://localhost
  https://remote.host/path/to/resource
param_list_ip =
  192.168.1.24
  10.0.0.3
param_list_network =
  192.168.3.0/24
  10.0.0.0/17
"""


class TestSettingsDefinition(unittest.TestCase):
    def test_valid_content(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        self.assertEqual(loader.name, "definition:yaml:raw")
        definition = SettingsDefinition(loader)
        self.assertTrue(definition.has_section("section1"))
        self.assertFalse(definition.has_section("unknown"))
        self.assertTrue(definition.section("section1").has_parameter("param_str"))
        self.assertFalse(definition.section("section1").has_parameter("unknown"))
        self.assertEqual(
            definition.section("section1").parameter("param_str").default, "value1"
        )
        self.assertEqual(
            definition.section("section2").parameter("param_int").example, 500
        )
        self.assertEqual(
            len(definition.section("section2").parameter("param_int").choices), 3
        )
        self.assertEqual(
            definition.section("section2").parameter("param_uri")._type, "uri"
        )
        self.assertIsNone(definition.section("section1").doc)
        self.assertIsNone(definition.section("section2").doc)

    def test_section_doc_valid(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["_doc"] = "Documentation for section2"
        definition = SettingsDefinition(loader)
        self.assertEqual(
            definition.section("section2").doc, "Documentation for section2"
        )
        self.assertEqual(
            definition.section("section2").parameter("param_int").default, 10
        )

    def test_section_doc_invalid_type_int(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["_doc"] = 42
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Invalid type for section _doc in \[section2\]$",
        ):
            SettingsDefinition(loader)

    def test_section_doc_invalid_type_dict(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["_doc"] = {"type": "str", "default": "fail"}
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Invalid type for section _doc in \[section2\]$",
        ):
            SettingsDefinition(loader)

    def test_section_doc_with_parameter_named_doc(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["_doc"] = "Section documentation"
        loader.content["section2"]["doc"] = {"type": "str", "default": "value"}
        definition = SettingsDefinition(loader)
        self.assertEqual(definition.section("section2").doc, "Section documentation")
        self.assertTrue(definition.section("section2").has_parameter("doc"))
        self.assertEqual(
            definition.section("section2").parameter("doc").default, "value"
        )

    def test_invalid_yaml(self):
        with self.assertRaisesRegex(
            SettingsDefinitionError, r"^Invalid YAML settings definition: "
        ):
            SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION + "\n fail")
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^YAML scanner error: mapping values are not allowed here.*",
        ):
            SettingsDefinitionLoaderYaml(raw="fail: again: ")

    def test_unsupported_property(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_int"]["unknown"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unsupported properties found for \[section2\]>param_int: unknown$",
        ):
            SettingsDefinition(loader)

    def test_unsupported_type(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_int"]["type"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unsupported type fail for \[section2\]>param_int$",
        ):
            SettingsDefinition(loader)

    def test_required_not_bool(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_required"]["required"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Invalid boolean value of \[section2\]>param_required required property$",
        ):
            SettingsDefinition(loader)

    def test_default_valid_type_int(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_int"]["default"] = 10
        definition = SettingsDefinition(loader)
        self.assertEqual(
            definition.section("section2").parameter("param_int").default, 10
        )

    def test_default_valid_type_bool(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_bool"]["default"] = True
        definition = SettingsDefinition(loader)
        self.assertEqual(
            definition.section("section2").parameter("param_bool").default, True
        )

    def test_default_valid_type_ip(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_ip"]["default"] = "192.168.0.10"
        definition = SettingsDefinition(loader)
        self.assertEqual(
            definition.section("section2").parameter("param_ip").default,
            ipaddress.ip_address("192.168.0.10"),
        )

    def test_default_valid_type_network(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_network"]["default"] = "192.168.0.0/24"
        definition = SettingsDefinition(loader)
        self.assertEqual(
            definition.section("section2").parameter("param_network").default,
            ipaddress.ip_network("192.168.0.0/24"),
        )

    def test_default_invalid_type_int(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_int"]["default"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Default value fail for parameter \[section2\]>param_int has not the "
            r"expected type int$",
        ):
            SettingsDefinition(loader)

    def test_default_invalid_type_bool(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_bool"]["default"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Default value fail for parameter \[section2\]>param_bool has not the "
            r"expected type bool$",
        ):
            SettingsDefinition(loader)

    def test_default_invalid_type_ip(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_ip"]["default"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Invalid default ip address value for parameter \[section2\]>param_ip$",
        ):
            SettingsDefinition(loader)

    def test_default_invalid_type_network(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_network"]["default"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Invalid default ip network value for parameter "
            r"\[section2\]>param_network$",
        ):
            SettingsDefinition(loader)

    def test_default_invalid_choice(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_int"]["default"] = 12
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Default value 12 for parameter \[section2\]>param_int is not one of "
            r"possible choices \[10, 100, 500\]$",
        ):
            SettingsDefinition(loader)

    def test_default_invalid_type_list(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_list"]["default"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Default value fail for parameter \[section2\]>param_list is not a valid "
            r"list$",
        ):
            SettingsDefinition(loader)

    def test_type_list_without_content(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        del loader.content["section2"]["param_list"]["content"]
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^List content type for parameter \[section2\]>param_list must be "
            r"defined$",
        ):
            SettingsDefinition(loader)

    def test_type_list_invalid_content_type(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_list"]["content"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unsupported list content type fail for \[section2\]>param_list$",
        ):
            SettingsDefinition(loader)
        loader.content["section2"]["param_list"]["content"] = "list"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unsupported list content type list for \[section2\]>param_list$",
        ):
            SettingsDefinition(loader)

    def test_type_content_on_other_type(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_list"]["type"] = "str"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Content property is forbidden for parameter \[section2\]>param_list "
            r"with type str$",
        ):
            SettingsDefinition(loader)

    def test_type_list_invalid_default_type(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_list"]["content"] = "bool"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Default value value1 for parameter \[section2\]>param_list has not the "
            r"expected type bool$",
        ):
            SettingsDefinition(loader)

    def test_type_list_invalid_choice(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_list"]["choices"] = ["value1", "value3"]
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Default value value2 for parameter \[section2\]>param_list is not one "
            r"of possible choices \['value1', 'value3'\]$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_invalid_type(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_deprecated"]["deprecated"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unsupported format of \[section2\]>param_deprecated deprecated "
            r"property$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_missing_key(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        del loader.content["section2"]["param_deprecated"]["deprecated"]["parameter"]
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unexpected keys in \[section2\]>param_deprecated deprecated property$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_unexpected_key(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_deprecated"]["deprecated"]["fail"] = "hello"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unexpected keys in \[section2\]>param_deprecated deprecated property$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_unsupported_section_type(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_deprecated"]["deprecated"]["section"] = 0
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unexpected type of \[section2\]>param_deprecated deprecated section "
            r"property$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_unsupported_parameter_type(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_deprecated"]["deprecated"]["parameter"] = 0
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Unexpected type of \[section2\]>param_deprecated deprecated parameter "
            r"property$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_required(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_deprecated"]["required"] = True
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Deprecated parameter \[section2\]>param_deprecated cannot be required$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_default(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_deprecated"]["default"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Deprecated parameter \[section2\]>param_deprecated cannot have default "
            r"value$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_invalid_section_reference(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_deprecated"]["deprecated"]["section"] = "fail"
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Invalid deprecated reference on parameter "
            r"\[section2\]>param_deprecated$",
        ):
            SettingsDefinition(loader)

    def test_deprecated_invalid_parameter_reference(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        loader.content["section2"]["param_deprecated"]["deprecated"]["parameter"] = (
            "fail"
        )
        with self.assertRaisesRegex(
            SettingsDefinitionError,
            r"^Invalid deprecated reference on parameter "
            r"\[section2\]>param_deprecated$",
        ):
            SettingsDefinition(loader)


class TestRuntimeSettings(unittest.TestCase):
    def test_valid_settings(self):
        loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(loader)
        settings = RuntimeSettings(definition)
        self.assertEqual(settings.section1.param_str, "value1")
        self.assertEqual(settings.section2.param_int, 10)
        self.assertEqual(settings.section2.param_path, Path("/path/to/file"))
        self.assertIsInstance(settings.section2.param_uri, tuple)
        self.assertEqual(settings.section2.param_uri.scheme, "https")
        self.assertIsInstance(settings.section1, RuntimeSettingsSection)
        self.assertEqual(type(settings.section1).__name__, "RuntimeSettingsSection1")

    def test_site_overrides(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        self.assertEqual(site_loader.name, "site:ini:raw")
        settings.override(site_loader)
        self.assertEqual(settings.section1.param_str, "site_value1")
        self.assertEqual(settings.section2.param_path, Path("/site/path/to/file"))
        self.assertEqual(settings.section2.param_list, ["value3", "value5"])
        self.assertIsInstance(settings.section2.param_ip, ipaddress.IPv4Address)
        self.assertEqual(str(settings.section2.param_ip), "127.0.0.1")
        self.assertTrue(settings.section2.param_ip.is_loopback)
        self.assertIsInstance(settings.section2.param_network, ipaddress.IPv4Network)
        self.assertEqual(str(settings.section2.param_network), "127.0.0.0/24")
        self.assertTrue(settings.section2.param_network.is_loopback)

    def test_site_override_invalid_section(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["big"] = {}
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Section big loaded in settings overrides is not defined in "
            r"settings definition$",
        ):
            settings.override(site_loader)

    def test_site_override_invalid_parameter(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["unknown"] = "fail"
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Parameter unknown loaded in settings overrides is not defined in "
            r"section section2 of settings definition$",
        ):
            settings.override(site_loader)

    def test_site_override_invalid_type_int(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["param_int"] = "fail"
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Invalid integer value 'fail' for \[section2\]>param_int in site "
            r"overrides$",
        ):
            settings.override(site_loader)

    def test_site_override_invalid_type_float(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["param_float"] = "fail"
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Invalid float value 'fail' for \[section2\]>param_float in site "
            r"overrides$",
        ):
            settings.override(site_loader)

    def test_site_override_invalid_type_bool(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["param_bool"] = "fail"
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Invalid boolean value 'fail' for \[section2\]>param_bool in site "
            r"overrides$",
        ):
            settings.override(site_loader)

    def test_site_override_type_ip_ipv6(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["param_ip"] = "::1"
        settings.override(site_loader)
        self.assertIsInstance(settings.section2.param_ip, ipaddress.IPv6Address)
        self.assertEqual(str(settings.section2.param_ip), "::1")
        self.assertTrue(settings.section2.param_ip.is_loopback)

    def test_site_override_invalid_type_ip(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["param_ip"] = "fail"
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Invalid ip address value 'fail' for \[section2\]>param_ip in site "
            r"overrides$",
        ):
            settings.override(site_loader)

    def test_site_override_type_network_ipv6(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        addr = "fd1a:aa9b:7d8b:cfbe::/64"  # Use random private network for this test
        site_loader.content["section2"]["param_network"] = addr
        settings.override(site_loader)
        self.assertIsInstance(settings.section2.param_network, ipaddress.IPv6Network)
        self.assertEqual(str(settings.section2.param_network), addr)
        self.assertTrue(settings.section2.param_network.is_private)

    def test_site_override_invalid_type_network(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["param_network"] = "fail"
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Invalid ip network value 'fail' for \[section2\]>param_network in site "
            r"overrides$",
        ):
            settings.override(site_loader)

    def test_site_override_invalid_choice(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["param_int"] = "12"
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Value 12 for parameter \[section2\]>param_int in site overrides is not "
            r"one of possible choices \[10, 100, 500\]$",
        ):
            settings.override(site_loader)

    def test_site_override_undefined_required(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        del site_loader.content["section2"]["param_required"]
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Parameter \[section2\]>param_required is missing but required in "
            r"settings overrides$",
        ):
            settings.override(site_loader)

    def test_site_override_list_invalid_content_type(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        def_loader.content["section2"]["param_list"]["content"] = "int"
        def_loader.content["section2"]["param_list"]["default"] = [2, 5]
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Invalid integer value 'value3' for \[section2\]>param_list in site "
            r"overrides$",
        ):
            settings.override(site_loader)

    def test_site_override_list_invalid_choice(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        def_loader.content["section2"]["param_list"]["choices"] = ["value1", "value2"]
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        with self.assertRaisesRegex(
            SettingsOverrideError,
            r"^Value value3 for parameter \[section2\]>param_list in site overrides is "
            r"not one of possible choices \['value1', 'value2'\]$",
        ):
            settings.override(site_loader)

    def test_site_override_deprecated(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        del site_loader.content["section1"]["param_str"]
        site_loader.content["section2"]["param_deprecated"] = "value"

        with self.assertWarnsRegex(
            FutureWarning,
            r"^Parameter \[section2\]>param_deprecated is deprecated in favor of "
            r"\[section1\]>param_str$",
        ):
            settings.override(site_loader)
        self.assertEqual(settings.section2.param_deprecated, "value")

    def test_site_override_skip(self):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        site_loader.content["section2"]["param_deprecated"] = "value"

        with self.assertLogs("rfl", level="WARNING") as cm:
            settings.override(site_loader)
        self.assertEqual(
            cm.output,
            [
                "WARNING:rfl.settings:Ignoring deprecated parameter "
                "[section2]>param_deprecated since replacement parameter "
                "[section1]>param_str is also defined"
            ],
        )
        self.assertIsNone(settings.section2.param_deprecated)

    @patch("sys.stdout", new_callable=StringIO)
    def test_dump(self, mock_stdout):
        def_loader = SettingsDefinitionLoaderYaml(raw=VALID_DEFINITION)
        definition = SettingsDefinition(def_loader)
        settings = RuntimeSettings(definition)
        site_loader = RuntimeSettingsSiteLoaderIni(VALID_SITE)
        settings.override(site_loader)
        settings.dump()
        self.assertEqual(
            mock_stdout.getvalue(),
            "[section1]\n"
            "  param_str: site_value1 (site:ini:raw)\n"
            "[section2]\n"
            "  param_int: 10 (definition:yaml:raw)\n"
            "  param_path: /site/path/to/file (site:ini:raw)\n"
            "  param_uri: https://localhost:5900/resources (definition:yaml:raw)\n"
            "  param_float: ∅ (definition:yaml:raw)\n"
            "  param_bool: True (definition:yaml:raw)\n"
            "  param_list: value3, value5 (site:ini:raw)\n"
            "  param_required: required_value (site:ini:raw)\n"
            "  param_password: •••••• (site:ini:raw)\n"
            "  param_ip: 127.0.0.1 (site:ini:raw)\n"
            "  param_network: 127.0.0.0/24 (site:ini:raw)\n"
            "  param_list_uri: http://localhost, https://remote.host/path/to/resource "
            "(site:ini:raw)\n"
            "  param_list_ip: 192.168.1.24, 10.0.0.3 (site:ini:raw)\n"
            "  param_list_network: 192.168.3.0/24, 10.0.0.0/17 (site:ini:raw)\n"
            "  param_deprecated[🪦]: ∅ (definition:yaml:raw)\n",
        )



# Copyright (c) 2023 Rackslab
#
# This file is part of RFL.
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import configparser
import logging
import warnings
from pathlib import Path
from typing import Optional, Set, Tuple

from rfl.authentication.user import AuthenticatedUser
import yaml

from .errors import RBACPolicyDefinitionLoadError, RBACPolicyRolesLoadError


logger = logging.getLogger(__name__)

ANONYMOUS_ROLE = "anonymous"
ALL_MEMBER = "ALL"


class ActionMetadata:
    """Metadata describing an action in the policy definition."""

    def __init__(
        self,
        name: str,
        description: str,
        deprecated: bool = False,
        replaced_by: Optional[str] = None,
    ):
        self.name = name
        self.description = description
        self.deprecated = deprecated
        self.replaced_by = replaced_by

    def __repr__(self):
        return (
            f"ActionMetadata(name={self.name}, deprecated={self.deprecated}, "
            f"replaced_by={self.replaced_by})"
        )


class RBACPolicyRole:
    def __init__(self, name: str, members: Set[str], actions: Set[str]):
        self.name = name
        self.members = members
        self.actions = actions

    def __repr__(self):
        return f"{self.name} [ members: {self.members}, actions: {self.actions} ]"


class RBACPolicyDefinitionLoader:
    pass


class RBACPolicyDefinitionYAMLLoader(RBACPolicyDefinitionLoader):
    def __init__(self, path: Path = None, raw: str = None):
        try:
            if raw is not None:
                content = yaml.safe_load(raw)
            elif path is not None:
                try:
                    with open(path) as fh:
                        content = yaml.safe_load(fh)
                except FileNotFoundError:
                    raise RBACPolicyDefinitionLoadError(
                        f"Policy definition file {path} not found"
                    )
            else:
                raise RBACPolicyDefinitionLoadError(
                    "Either a raw string value or a path must be given to load YAML "
                    "policy definition"
                )
        except yaml.parser.ParserError as err:
            raise RBACPolicyDefinitionLoadError(
                f"Invalid YAML policy definition: {str(err)}"
            ) from err
        try:
            actions_content = content["actions"]
        except KeyError as err:
            raise RBACPolicyDefinitionLoadError(
                "Actions key not found in YAML policy definition"
            ) from err
        except AttributeError as err:
            raise RBACPolicyDefinitionLoadError(
                "Unable to extract the set of actions from actions key in YAML policy "
                "definition"
            ) from err
        if not isinstance(actions_content, dict):
            raise RBACPolicyDefinitionLoadError(
                "Actions definition must be a mapping of action names to definitions"
            )

        self.actions = {}
        for action, definition in actions_content.items():
            if isinstance(definition, str):
                description = definition
                deprecated = False
                replaced_by = None
            elif isinstance(definition, dict):
                try:
                    description = definition["description"]
                except KeyError as err:
                    raise RBACPolicyDefinitionLoadError(
                        f"Description is mandatory for action {action}"
                    ) from err
                deprecated = definition.get("deprecated", False)
                if not isinstance(deprecated, bool):
                    raise RBACPolicyDefinitionLoadError(
                        f"Invalid deprecated value for action {action}"
                    )
                replaced_by = definition.get("replaced_by")
                if replaced_by is not None and not isinstance(replaced_by, str):
                    raise RBACPolicyDefinitionLoadError(
                        f"Invalid replaced_by value for action {action}"
                    )
            else:
                raise RBACPolicyDefinitionLoadError(
                    f"Invalid definition for action {action}"
                )
            self.actions[action] = ActionMetadata(
                name=action,
                description=description,
                deprecated=deprecated,
                replaced_by=replaced_by,
            )

        # Validate replacement references
        for action, metadata in self.actions.items():
            if metadata.replaced_by and metadata.replaced_by not in self.actions:
                raise RBACPolicyDefinitionLoadError(
                    f"Replacement action {metadata.replaced_by} for {action} not found "
                    "in policy definition"
                )

    def apply(self, action: str) -> Set[str]:
        """Return effective actions for the given action name, handling deprecation.

        Deprecated actions are stripped, a UserWarning is emitted, and a single
        replacement action (if declared) is added instead. Unknown actions still raise
        RBACPolicyRolesLoadError.
        """
        metadata = self.actions[action]
        if metadata.deprecated:
            warning_msg = f"Action {action} is deprecated"
            if metadata.replaced_by:
                warning_msg += f"; use {metadata.replaced_by} instead"
            warnings.warn(warning_msg, UserWarning)
            if metadata.replaced_by:
                return self.apply(metadata.replaced_by)
            return set()

        return {action}


class RBACPolicyRolesLoader:
    pass


class RBACPolicyRolesIniLoader(RBACPolicyRolesLoader):
    def __init__(
        self,
        definition: RBACPolicyDefinitionLoader,
        path: Path = None,
        raw: str = None,
        preload: bool = True,
    ):
        self.definition = definition
        self.content = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation(),
            allow_no_value=True,
        )
        try:
            if raw is not None:
                logger.debug("Loading settings in INI format from raw value")
                self.content.read_string(raw)
            elif path is not None:
                logger.debug("Loading settings file %s", path)
                self.content.read(path)
            else:
                raise RBACPolicyRolesLoadError(
                    "Either a raw string value or a path must be given to load site "
                    "settings"
                )
        except configparser.Error as err:
            raise RBACPolicyDefinitionLoadError(str(err)) from err
        self.roles = set()
        if preload:
            self._load()

    def _load(self):
        """Load the set of PolicyRoles defined in file pointed by the given path."""
        try:
            roles = self.content.options("roles")
        except configparser.NoSectionError as err:
            raise RBACPolicyRolesLoadError(
                "Section roles to define roles members is not defined in INI content"
            ) from err
        for role in roles:
            try:
                actions = self._expand_actions(self.content.get(role, "actions"))
            except configparser.NoSectionError as err:
                raise RBACPolicyRolesLoadError(
                    f"Section to define actions for role {role} is not defined in INI "
                    "content"
                ) from err
            except configparser.NoOptionError as err:
                raise RBACPolicyRolesLoadError(
                    f"Option to define actions for role {role} is not defined in INI "
                    "content"
                ) from err
            if role == ANONYMOUS_ROLE:
                members = None
            else:
                members = self._expand_members(self.content.get("roles", role))
            self.roles.add(RBACPolicyRole(role, members, actions))

    def _expand_actions(self, actions_str: str) -> Set[str]:
        """Return the set of actions declared in comma-separated list provided in
        actions_str argument. If an item is prefixed by @, the set is expanded with the
        actions of the role name that follows."""
        actions = set()
        for action in actions_str.split(","):
            if action.startswith("@"):
                actions.update(self._role_actions(action[1:]))
            else:
                if action not in self.definition.actions:
                    raise RBACPolicyRolesLoadError(
                        f"Action {action} not found in policy definition"
                    )
                actions.update(self.definition.apply(action))
        return actions

    def _expand_members(self, members_str):
        """Return the set of members declared in comma-separated list provided in
        members_str argument."""
        return set(members_str.split(","))

    def _role_actions(self, role):
        """Return the set of actions allowed the to given role. Raise
        RBACPolicyVerificationError if the role is not found in policy."""
        for _role in self.roles:
            if _role.name == role:
                return _role.actions

        raise RBACPolicyRolesLoadError(
            f"Unable to inherit actions from role {role} not found in policy"
        )


class RBACPolicyManager:
    def __init__(self, loader: RBACPolicyRolesLoader):
        self.loader = loader

    @property
    def allow_anonymous(self) -> bool:
        """Return True if the anonymous role is declared in policy, False otherwise."""
        for role in self.loader.roles:
            if role.name == ANONYMOUS_ROLE:
                return True
        return False

    def disable_anonymous(self) -> None:
        """Ensure anonymous role is disabled, even if defined in loaded policy."""
        for role in self.loader.roles.copy():
            if role.name == ANONYMOUS_ROLE:
                self.loader.roles.remove(role)

    def _user_roles(self, user: AuthenticatedUser) -> Set[str]:
        """Return the set of roles associated to a given user name."""
        roles = set()
        assert not user.is_anonymous()
        for role in self.loader.roles:
            # Skip anonymous role w/o members
            if role.name == ANONYMOUS_ROLE:
                continue
            # Include role with ALL authenticated members
            if ALL_MEMBER in role.members:
                roles.add(role)
            else:
                for member in role.members:
                    if member == user.login or (
                        member.startswith("@") and member[1:] in user.groups
                    ):
                        roles.add(role)
        logger.debug("Found the following roles for user %s: %s", user, roles)
        return roles

    def _anonymous_actions(self) -> Tuple[Set[str], Set[str]]:
        """Return tuple with set of role names and set of allowed actions for
        anonymous user."""
        for role in self.loader.roles:
            if role.name == ANONYMOUS_ROLE:
                return ({role.name}, role.actions)
        return (set(), set())

    def _user_actions(self, user: AuthenticatedUser) -> Tuple[Set[str], Set[str]]:
        """Return tuple with set of role names and set of allowed actions for a
        particular user and list of groups membership."""
        roles = self._user_roles(user)
        actions = set()
        for role in roles:
            actions.update(role.actions)
        return (set([role.name for role in roles]), actions)

    def roles_actions(self, user: AuthenticatedUser) -> Tuple[Set[str], Set[str]]:
        """Return tuple with set of role names and set of allowed actions for a
        given user."""
        if user.is_anonymous():
            return self._anonymous_actions()
        return self._user_actions(user)

    def allowed_anonymous_action(self, action: str) -> bool:
        """Return True if the given action is allowed for anonymous role, False
        otherwise."""
        for role in self.loader.roles:
            if role.name == ANONYMOUS_ROLE and action in role.actions:
                return True
        return False

    def allowed_user_action(self, user: AuthenticatedUser, action: str) -> bool:
        """Return True if the given action is allowed for the given user, False
        otherwise."""
        if user.is_anonymous():
            return self.allowed_anonymous_action(action)
        for role in self._user_roles(user):
            if action in role.actions:
                logger.debug(
                    "Token for user %s is permitted to perform action %s",
                    user,
                    action,
                )
                return True
        logger.warning(
            "Token for user %s is not authorized to perform action %s",
            user,
            action,
        )
        return False

    @classmethod
    def yaml_definition_ini_roles(cls, definition: Path, roles: Path):
        return cls(
            RBACPolicyRolesIniLoader(
                definition=RBACPolicyDefinitionYAMLLoader(path=definition), path=roles
            )
        )
